import json
import logging
from typing import Any, Dict, List, Union

from jsonpath_ng import Index, Fields
from jsonpath_ng.ext import parse

from robot.api.deco import keyword


JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class JSON:
    r"""`JSON` is a library for manipulating `JSON`_ files and strings.

    JSON is a common data interchange format inspired by a subset of
    the Javascript programming language, but these days is a de facto
    standard in modern web APIs and is language agnostic.

    .. _JSON: http://json.org/

    **Serialization**

    The term `serialization` refers to the process of converting
    Robot Framework or Python types to JSON or the other way around.

    Basic types can be easily converted between the domains,
    and the mapping is as follows:

    ============= =======
    JSON          Python
    ============= =======
    object        dict
    array         list
    string        str
    number (int)  int
    number (real) float
    true          True
    false         False
    null          None
    ============= =======

    **About JSONPath**

    Reading and writing values from/to JSON serializable objects is done
    using `JSONPath`_. It's a syntax designed to quickly and easily refer to
    specific elements in a JSON structure.

    Compared to Python's normal dictionary access, JSONPath expressions can
    target multiple elements through features such as conditionals and wildcards,
    which can simplify many JSON-related operations. It's analogous to XPath
    for XML structures.

    .. _JSONPath: http://goessner.net/articles/JsonPath/

    **Syntax example**

    For this example consider the following structure:

    .. code-block:: json

       {
         "clients": [
           {
             "name": "Johnny Example",
             "email": "john@example.com",
             "orders": [
                 {"address": "Streetroad 123", "price": 103.20},
                 {"address": "Streetroad 123", "price": 98.99}
             ]
           },
           {
             "name": "Jane Example",
             "email": "jane@example.com",
             "orders": [
                 {"address": "Waypath 321", "price": 22.00},
                 {"address": "Streetroad 123", "price": 2330.01}
             ]
           }
         ]
       }

    In the simplest case JSONPath can replace nested access:

    .. code-block:: robotframework

       # First order of first client, with direct dictionary access
       ${value}=    Set variable    ${json}["clients"][0]["orders"][0]

       # JSONPath access
       ${value}=    Get value from JSON    ${json}    $.clients[0].orders[0]

    But the power comes from complicated expressions:

    .. code-block:: robotframework

       # Find delivery addresses for all orders
       ${prices}=        Get values from JSON    $..address

       # Find orders that cost over 100
       ${expensives}=    Get values from JSON    $..orders[?(@.price>100)]


    **Supported Expressions**

    The supported syntax elements are:

    ======== ===========
    Element  Description
    ======== ===========
    $        Root object/element
    @        Current object/element
    \. or [] Child operator
    \.\.     Recursive descent
    \*       Wilcard, any element
    [n]      Array index
    [a:b:c]  Array slice (start, end, step)
    [a,b]    Union of indices or names
    ?()      Apply a filter expression
    ()       Script expression
    ======== ===========

    There are a multitude of different script expressions
    in addition to the elements listed above, which can
    be seen in the `aforementioned article`__.

    For further library usage examples, see the individual keywords.

    __ JSONPath_
    """

    # TODO: Add more logging about affected rows, at least on debug level

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @keyword("Load JSON from file")
    def load_json_from_file(self, filename: str) -> JSONType:
        """Load JSON data from a file, and return it as JSON serializable object.
        Depending on the input file the object can be either a dictionary,
        a list, or a scalar value.

        :param filename: path to input file

        Example:

        .. code:: robotframework

           &{auth}=    Load JSON from file    auth.json
           Log   Current auth token: ${auth.token}

        """
        self.logger.info("Loading JSON from file: %s", filename)
        with open(filename) as json_file:
            return json.load(json_file)

    @keyword("Save JSON to file")
    def save_json_to_file(self, doc: JSONType, filename: str):
        """Save a JSON serializable object or a string containg
        a JSON value into a file.

        :param doc: JSON serializable object or string
        :param filename: path to output file

        Example:

        .. code:: robotframework

           # Save dictionary to file
           ${john}=    Create dictionary    name=John    mail=john@example.com
           Save JSON to file    ${john}    john.json

           # Save string to file
           ${mark}=    Set variable    {"name": "Mark", "mail": "mark@example.com"}
           Save JSON to file    ${mark}    mark.json

        """
        self.logger.info("Saving JSON to file: %s", filename)
        doc = self.convert_string_to_json(doc) if isinstance(doc, str) else doc
        with open(filename, "w") as outfile:
            json.dump(doc, outfile)

    @keyword("Convert JSON to String")
    def convert_json_to_string(self, doc: JSONType) -> str:
        """Convert a JSON serializable object to a string and return it.

        :param doc: JSON serializable object

        Example:

        .. code:: robotframework

           ${obj}=    Create dictionary    Key=Value
           ${json}=   Convert JSON to string    ${obj}
           Should be equal    ${json}     {"Key": "Value"}

        """
        return json.dumps(doc)

    @keyword("Convert String to JSON")
    def convert_string_to_json(self, doc: str) -> JSONType:
        """Convert a string to a JSON serializable object and return it.

        :param doc: JSON string

        Example:

        .. code:: robotframework

           ${json}=    Set variable    {"Key": "Value"}
           &{obj}=     Convert string to JSON    ${json}
           Should be equal    ${obj.Key}    Value

        """
        return json.loads(doc)

    @keyword("Add to JSON")
    def add_to_json(self, doc: JSONType, expr: str, value: JSONType):
        """Add items into a JSON serializable object and return the result.

        If the target is a list, the values are appended to the end.
        If the target is a dict, the keys are either added or updated.

        :param doc: JSON serializable object
        :param expr: JSONPath expression
        :param value: values to either append or update

        Example:

        .. code:: robotframework

           # Change the name value for all people
           &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
           &{after}=     Add to JSON    ${json}   $.People.name    JohnMalkovich

        """  # noqa: E501
        self.logger.info('Add to JSON with expression: "%s"', expr)
        for match in parse(expr).find(doc):
            if isinstance(match.value, dict):
                match.value.update(value)
            if isinstance(match.value, list):
                match.value.append(value)
        return doc

    @keyword("Get value from JSON")
    def get_value_from_json(self, doc: JSONType, expr: str, default: Any = None):
        """Get a single value from a JSON serializable object that matches the given expression.

        Raises a ValueError if there is more than one match.
        Returns the given default argument (or None) if there
        were no matches.

        :param doc: JSON serializable object or string
        :param expr: jsonpath expression

        Example:

        .. code:: robotframework

           # Get the name value for the first person
           &{people}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
           ${first}=     Get value from JSON      ${people}   $.People[0].name

        """  # noqa: E501
        self.logger.info('Get value from JSON with expression: "%s"', expr)
        result = [match.value for match in parse(expr).find(doc)]
        if len(result) == 0:
            return default
        elif len(result) == 1:
            return result[0]
        else:
            raise ValueError(
                "Found {count} matches: {values}".format(
                    count=len(result), values=", ".join(str(r) for r in result)
                )
            )

    @keyword("Get values from JSON")
    def get_values_from_json(self, doc: JSONType, expr: str):
        """Get all values from a JSON serializable object that match the given expression.

        :param doc: JSON serializable object or string
        :param expr: JSONPath expression

        Example:

        .. code:: robotframework

           # Get all the names for all people
           &{people}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
           @{names}=     Get values from JSON     ${people}   $.People[*].name

        """  # noqa: E501
        self.logger.info('Get values from JSON with expression: "%s"', expr)
        return [match.value for match in parse(expr).find(doc)]

    @keyword("Update value to JSON")
    def update_value_to_json(self, doc: JSONType, expr: str, value: JSONType):
        """Update existing values in a JSON serializable object and return the result.
        Will change all values that match the expression.

        :param doc: JSON or string
        :param expr: JSONPath expression
        :param value: New value for the matching item(s)

        Example:

        .. code:: robotframework

           # Change the name key for all people
           &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
           &{after}=     Update value to JSON     ${json}   $.People[*].name    JohnMalkovich

        """  # noqa: E501
        self.logger.info('Update JSON with expression: "%s"', expr)
        for match in parse(expr).find(doc):
            path = match.path
            if isinstance(path, Index):
                match.context.value[match.path.index] = value
            elif isinstance(path, Fields):
                match.context.value[match.path.fields[0]] = value
        return doc

    @keyword("Delete from JSON")
    def delete_from_json(self, doc: JSONType, expr: str):
        """Delete values from a JSON serializable object and return the result.
        Will delete all values that match the expression.

        :param doc: JSON serializable object or string
        :param expr: JSONPath expression

        Example:

        .. code:: robotframework

           # Delete all people
           &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
           &{after}=     Delete from JSON    ${json}   $.People[*]
        """  # noqa: E501
        self.logger.info('Delete from JSON with expression: "%s"', expr)
        for match in parse(expr).find(doc):
            path = match.path
            if isinstance(path, Index):
                del match.context.value[match.path.index]
            elif isinstance(path, Fields):
                del match.context.value[match.path.fields[0]]
        return doc



import base64
from enum import Enum
from functools import wraps
from io import BytesIO
import json
import logging
import os
from pathlib import Path
import pickle
import shutil
import tempfile
import time
from typing import Any


try:
    from apiclient import discovery  # MediaIoBaseDownload
    from apiclient.errors import HttpError
    from apiclient.http import MediaFileUpload, MediaIoBaseDownload
    from googleapiclient.discovery import build
    from google_auth_oauthlib.flow import InstalledAppFlow
    from google.auth.transport.requests import Request
    from google.cloud import vision
    from google.cloud import language_v1
    from google.cloud.language_v1 import enums
    from google.cloud import storage
    from google.cloud import videointelligence
    from google.cloud import translate_v3
    from google.cloud import texttospeech_v1
    from google.cloud.texttospeech_v1.types import (
        AudioConfig,
        VoiceSelectionParams,
        SynthesisInput,
    )
    from google.cloud import speech
    from google.oauth2 import service_account
    from google.protobuf.json_format import MessageToJson

    GOOGLECLOUD_IMPORT_ERROR = None
except ImportError as e:
    GOOGLECLOUD_IMPORT_ERROR = str(e)

from RPA.Robocloud.Secrets import Secrets


def google_dependency_required(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if GOOGLECLOUD_IMPORT_ERROR:
            raise ValueError(
                "Please install optional `google` package, "
                "`pip install rpaframework[google]` to use RPA.Cloud.Google library\n"
                + GOOGLECLOUD_IMPORT_ERROR
            )
        return f(*args, **kwargs)

    return wrapper


class GoogleOAuthAuthenticationError(Exception):
    """Raised when unable to get Google OAuth credentials."""


class GoogleDriveError(Exception):
    """Raised with errors in Drive API"""


class Update(Enum):
    """Possible file update actions."""

    trash = 1
    untrash = 2
    star = 3
    unstar = 4


def to_action(value):
    """Convert value to Update enum."""
    if isinstance(value, Update):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Update[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown file update action: {value}") from err


class GoogleBase:
    """Google base class for generic methods"""

    logger = None
    services: list = []
    clients: dict = {}
    region: str = None
    robocloud_vault_name: str = None
    robocloud_vault_secret_key: str = None
    global_scopes: list = []
    _scopes: list = []

    def _get_service(self, service_name: str = None):
        """Return client instance for servive if it has been initialized.

        :param service_name: name of the Google service
        :return: client instance
        """
        if service_name not in self.clients.keys():
            raise KeyError(
                "Google service %s has not been initialized" % service_name.upper()
            )
        return self.clients[service_name]

    def _set_service(self, service_name: str = None, service: Any = None):
        self.clients[service_name] = service

    @google_dependency_required
    def _write_json(self, json_file, response):
        if json_file and response:
            with open(json_file, "w") as f:
                f.write(MessageToJson(response))

    def _get_service_account_from_robocloud(self):
        temp_filedesc = None
        if self.robocloud_vault_name is None or self.robocloud_vault_secret_key is None:
            raise KeyError(
                "Both 'robocloud_vault_name' and 'robocloud_vault_secret_key' "
                "are required to access Robocloud Vault. Set them in library "
                "init or with `set_robocloud_vault` keyword."
            )
        vault = Secrets()

        vault_items = vault.get_secret(self.robocloud_vault_name)
        secret = json.loads(vault_items[self.robocloud_vault_secret_key].strip())
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_filedesc:
            json.dump(secret, temp_filedesc, ensure_ascii=False)

        return temp_filedesc.name

    def _get_filecontent_from_robocloud(
        self, vault_name: str = None, secret_key: str = None
    ):
        vault_name = vault_name or self.robocloud_vault_name
        secret_key = secret_key or self.robocloud_vault_secret_key
        if vault_name is None or secret_key is None:
            raise KeyError(
                "Both 'robocloud_vault_name' and 'robocloud_vault_secret_key' "
                "are required to access Robocloud Vault. Set them in library "
                "init or with `set_robocloud_vault` keyword."
            )
        vault = Secrets()
        vault_items = vault.get_secret(vault_name)
        if isinstance(vault_items[secret_key], dict):
            return vault_items[secret_key]
        else:
            return vault_items[secret_key].strip()

    @google_dependency_required
    def _init_with_robocloud(self, client_object, service_name):
        temp_filedesc = self._get_service_account_from_robocloud()
        try:
            service = client_object.from_service_account_json(temp_filedesc)
            self._set_service(service_name, service)
        finally:
            if temp_filedesc:
                os.remove(temp_filedesc)

    def _init_service(
        self, client_object, service_name, service_credentials_file, use_robocloud_vault
    ):
        if use_robocloud_vault:
            self._init_with_robocloud(client_object, service_name)
        elif service_credentials_file:
            service = client_object.from_service_account_json(service_credentials_file)
            self._set_service(service_name, service)
        else:
            service = client_object()
            self._set_service(service_name, service)

    def set_robocloud_vault(self, vault_name: str = None, vault_secret_key: str = None):
        """Set Robocloud Vault name and secret key name

        :param vault_name: Robocloud Vault name
        :param vault_secret_key: Rococloud Vault secret key name
        """
        if vault_name:
            self.robocloud_vault_name = vault_name
        if vault_secret_key:
            self.robocloud_vault_secret_key = vault_secret_key

    def set_global_scopes(self, scopes: list = None):
        """Set global Google authentication scopes

        Useful when using numerous services with different scopes

        :param scopes: list of authentication scopes
        """
        if isinstance(scopes, list):
            self.global_scopes = scopes
        else:
            raise AttributeError("scopes needs to be a list")

    def _get_credentials_with_oauth_token(
        self, use_robocloud_vault, token_file, credentials_file, scopes, save_token
    ):
        credentials = None
        token_file_location = Path(token_file).absolute()
        if use_robocloud_vault:
            token = self._get_filecontent_from_robocloud(secret_key="oauth-token")
            credentials = pickle.loads(base64.b64decode(token))
        else:
            if os.path.exists(token_file_location):
                with open(token_file_location, "rb") as token:
                    credentials = pickle.loads(token)
        if scopes:
            self._scopes = self._scopes + [
                f"https://www.googleapis.com/auth/{scope}" for scope in scopes
            ]

        if not credentials or not credentials.valid:
            if credentials and credentials.expired and credentials.refresh_token:
                credentials.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    credentials_file, self._scopes
                )
                credentials = flow.run_local_server()
            if save_token:
                with open(token_file_location, "wb") as token:
                    pickle.dump(credentials, token)
        if not credentials:
            raise GoogleOAuthAuthenticationError(
                "Could not get Google OAuth credentials"
            )
        return credentials


class ServiceVision(GoogleBase):
    """Class for Google Cloud Vision API

    Link to `Vision PyPI`_ page.

    .. _Vision PyPI: https://pypi.org/project/google-cloud-vision/
    """

    _service_name = "vision"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceVision init")

    @google_dependency_required
    def init_vision_client(
        self,
        service_credentials_file: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Cloud Vision client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            vision.ImageAnnotatorClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def _get_google_image(self, image_file):
        if not image_file:
            raise KeyError("image_file is required for parameter")
        with open(image_file, "rb") as f:
            content = f.read()
        return vision.types.Image(content=content)  # pylint: disable=E1101

    @google_dependency_required
    def detect_labels(self, image_file: str, json_file: str = None) -> dict:
        """Detect labels in the image

        :param image_file: source image file
        :param json_file: json target to save result, defaults to None
        :return: detection response
        """
        service = self._get_service(self._service_name)
        image = self._get_google_image(image_file)
        response = service.label_detection(image=image)
        self._write_json(json_file, response)
        return response

    @google_dependency_required
    def detect_text(self, image_file: str, json_file: str = None) -> dict:
        """Detect text in the image

        :param image_file: source image file
        :param json_file: json target to save result, defaults to None
        :return: detection response
        """
        service = self._get_service(self._service_name)
        image = self._get_google_image(image_file)
        response = service.text_detection(image=image)
        self._write_json(json_file, response)
        return response

    @google_dependency_required
    def detect_document(self, image_file: str, json_file: str = None) -> dict:
        """Detect document

        :param image_file: source image file
        :param json_file: json target to save result, defaults to None
        :return: detection response
        """
        service = self._get_service(self._service_name)
        image = self._get_google_image(image_file)
        response = service.document_text_detection(image=image)
        self._write_json(json_file, response)
        return response

    @google_dependency_required
    def annotate_image(self, image_uri: str, json_file: str = None) -> dict:
        """Annotate image

        :param image_file: source image file
        :param json_file: json target to save result, defaults to None
        :return: detection response
        """
        service = self._get_service(self._service_name)
        response = service.annotate_image(
            {"image": {"source": {"image_uri": image_uri}}}
        )
        self._write_json(json_file, response)
        return response

    @google_dependency_required
    def face_detection(self, image_uri: str, json_file: str = None) -> dict:
        """Detect faces

        :param image_uri: Google Cloud Storage URI
        :param json_file: json target to save result, defaults to None
        :return: detection response
        """
        service = self._get_service(self._service_name)
        response = service.face_detection({"source": {"image_uri": image_uri}})
        self._write_json(json_file, response)
        return response


class ServiceNaturalLanguage(GoogleBase):
    """Class for Google Cloud Natural Language API

    Link to `Natural Language PyPI`_ page.

    .. _Natural Language PyPI: https://pypi.org/project/google-cloud-language/
    """

    _service_name = "natural-language"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceNaturalLanguage init")

    @google_dependency_required
    def init_natural_language_client(
        self, service_credentials_file: str = None, use_robocloud_vault: bool = False
    ) -> None:
        """Initialize Google Cloud Natural Language client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            language_v1.LanguageServiceClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def analyze_sentiment(
        self, text_file: str, file_type: str = "text", json_file: str = None, lang=None
    ) -> dict:
        """Analyze sentiment in a text file

        :param text_file: source text file
        :param json_file: json target to save result, defaults to None
        :param lang: language code of the source, defaults to None
        :return: analysis response
        """
        service = self._get_service(self._service_name)
        with open(text_file, "r") as f:
            text_content = f.read()
        # Available types: PLAIN_TEXT, HTML
        if file_type == "text":
            type_ = enums.Document.Type.PLAIN_TEXT
        else:
            type_ = enums.Document.Type.HTML
        document = {"content": text_content, "type": type_}
        # Optional. If not specified, the language is automatically detected.
        # For list of supported languages:
        # https://cloud.google.com/natural-language/docs/languages
        if lang is not None:
            document["language"] = lang
        # Available values: NONE, UTF8, UTF16, UTF32
        encoding_type = enums.EncodingType.UTF8
        response = service.analyze_sentiment(document, encoding_type=encoding_type)
        self._write_json(json_file, response)
        return response

    @google_dependency_required
    def classify_text(self, text_file, json_file, lang=None):
        """Classify text

        :param text_file: source text file
        :param json_file: json target to save result, defaults to None
        :param lang: language code of the source, defaults to None
        :return: classify response
        """
        service = self._get_service(self._service_name)
        with open(text_file, "r") as f:
            text_content = f.read()
        # Available types: PLAIN_TEXT, HTML
        type_ = enums.Document.Type.PLAIN_TEXT
        document = {"content": text_content, "type": type_}
        # Optional. If not specified, the language is automatically detected.
        # For list of supported languages:
        # https://cloud.google.com/natural-language/docs/languages
        if lang is not None:
            document["language"] = lang
        response = service.classify_text(document)
        self._write_json(json_file, response)
        return response


class ServiceVideoIntelligence(GoogleBase):
    """Class for Google Cloud Video Intelligence API

    API supports only `Google Cloud Storages`_ URIs at the moment.

    Link to `Video Intelligence PyPI`_ page.

    .. _Video Intelligence PyPI: https://pypi.org/project/google-cloud-videointelligence
    .. _Google Cloud Storages: https://cloud.google.com/storage/
    """

    _service_name = "video-intelligence"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceVideoIntelligence init")

    @google_dependency_required
    def init_video_intelligence_client(
        self,
        service_credentials_file: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Cloud Video Intelligence client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            videointelligence.VideoIntelligenceServiceClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def annotate_video(
        self,
        video_uri: str = None,
        video_file: str = None,
        json_file: str = None,
        features: list = None,
    ):
        """Annotate video

        Possible values for features:

        - FEATURE_UNSPECIFIED, Unspecified.
        - LABEL_DETECTION, Label detection. Detect objects, such as dog or flower.
        - SHOT_CHANGE_DETECTION, Shot change detection.
        - EXPLICIT_CONTENT_DETECTION, Explicit content detection.
        - SPEECH_TRANSCRIPTION, Speech transcription.
        - TEXT_DETECTION, OCR text detection and tracking.
        - OBJECT_TRACKING, Object detection and tracking.
        - LOGO_RECOGNITION, Logo detection, tracking, and recognition.

        If `video_uri` is given then that is used even if `video_file` is None.

        :param video_uri: Google Cloud Storage URI
        :param video_file: filepath to video
        :param json_file: json target to save result, defaults to None
        :param features: list of annotation features to detect,
            defaults to ["LABEL_DETECTION", "SHOT_CHANGE_DETECTION"]
        :return: annotate result
        """
        service = self._get_service(self._service_name)
        response = None
        if features is None:
            features = ["LABEL_DETECTION", "SHOT_CHANGE_DETECTION"]
        if video_uri:
            response = service.annotate_video(
                input_uri=video_uri, features=features
            ).result()
        elif video_file:
            with open(video_file, "rb") as f:
                response = service.annotate_video(
                    input_content=f.read(), features=features
                ).result()
        self._write_json(json_file, response)
        return response


class ServiceTranslation(GoogleBase):
    """Class for Google Cloud Translation API

    Link to `Translation PyPI`_ page.

    .. _Translation PyPI: https://pypi.org/project/google-cloud-translate/
    """

    _service_name = "translation"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceTranslation init")
        self._project_id = ""

    @google_dependency_required
    def init_translation_client(
        self,
        service_credentials_file: str = None,
        project_identifier: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Cloud Translation client

        :param service_credentials_file: filepath to credentials JSON
        :param project_identifier: identifier for Translation project
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            translate_v3.TranslationServiceClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )
        self._project_id = project_identifier

    @google_dependency_required
    def translate(
        self, text: Any, source_language: str = None, target_language: str = None
    ) -> dict:
        """Translate text

        :param text: text to translate
        :param source_language: language code, defaults to None
        :param target_language: language code, defaults to None
        :return: translated text
        """
        service = self._get_service(self._service_name)
        if not text and not target_language:
            raise KeyError("text and target_language are required parameters")
        parent = service.location_path(self._project_id, "global")
        if isinstance(text, str):
            text = [text]
        response = service.translate_text(
            contents=text,
            source_language_code=source_language,
            target_language_code=target_language,
            parent=parent,
        )
        return response


class ServiceTextToSpeech(GoogleBase):
    """Class for Google Cloud Text-to-Speech API

    Link to `Text To Speech PyPI`_ page.

    .. _Text To Speech PyPI: https://pypi.org/project/google-cloud-texttospeech/
    """

    _service_name = "text-to-speech"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceTextToSpeech init")

    @google_dependency_required
    def init_text_to_speech_client(
        self, service_credentials_file: str = None, use_robocloud_vault: bool = False
    ) -> None:
        """Initialize Google Cloud Text to Speech client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            texttospeech_v1.TextToSpeechClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def list_supported_voices(self, language_code: str = None):
        """List supported voices for the speech

        :param language_code: voice languages to list, defaults to None (all)
        :return: list of supported voices
        """
        service = self._get_service(self._service_name)
        if language_code:
            voices = service.list_voices(language_code)
        else:
            voices = service.list_voices()
        return voices.voices

    @google_dependency_required
    def synthesize_speech(
        self,
        text,
        language="en-US",
        name="en-US-Standard-B",
        gender="MALE",
        encoding="MP3",
        target_file="synthesized.mp3",
    ):
        """Synthesize speech synchronously

        :param text: input text to synthesize
        :param language: voice language, defaults to "en-US"
        :param name: voice name, defaults to "en-US-Standard-B"
        :param gender: voice gender, defaults to "MALE"
        :param encoding: result encoding type, defaults to "MP3"
        :param target_file: save synthesized output to file,
            defaults to "synthesized.mp3"
        :return: synthesized output in bytes
        """
        if not text:
            raise KeyError("text is required for kw: synthesize_speech")
        service = self._get_service(self._service_name)
        synth_input = SynthesisInput(text=text)
        voice_selection = VoiceSelectionParams(
            language_code=language, name=name, ssml_gender=gender
        )
        audio_config = AudioConfig(audio_encoding=encoding)
        response = service.synthesize_speech(synth_input, voice_selection, audio_config)
        if target_file:
            with open(target_file, "wb") as f:
                f.write(response.audio_content)
        return response.audio_content


class ServiceSpeechToText(GoogleBase):
    """Class for Google Cloud Speech-To-Text API

    Possible input audio encodings:

    - 'AMR'
    - 'AMR_WB'
    - 'ENCODING_UNSPECIFIED'
    - 'FLAC'
    - 'LINEAR16'
    - 'MULAW'
    - 'OGG_OPUS'
    - 'SPEEX_WITH_HEADER_BYTE'

    Link to `Speech To Text PyPI`_ page.

    .. _Speech To Text PyPI: https://pypi.org/project/google-cloud-speech/
    """

    _service_name = "speech-to-text"
    if not GOOGLECLOUD_IMPORT_ERROR:
        _encodings = {
            "AMR": speech.enums.RecognitionConfig.AudioEncoding.AMR,
            "AMR_WB": speech.enums.RecognitionConfig.AudioEncoding.AMR_WB,
            "FLAC": speech.enums.RecognitionConfig.AudioEncoding.FLAC,
            "LINEAR16": speech.enums.RecognitionConfig.AudioEncoding.LINEAR16,
            "MULAW": speech.enums.RecognitionConfig.AudioEncoding.MULAW,
            "OGG": speech.enums.RecognitionConfig.AudioEncoding.OGG_OPUS,
            "SPEEX": speech.enums.RecognitionConfig.AudioEncoding.SPEEX_WITH_HEADER_BYTE,  # noqa: E501 # pylint: disable=C0301
            "UNSPECIFIED": speech.enums.RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED,  # noqa: E501 # pylint: disable=C0301
        }
    else:
        _encodings = {}

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceSpeechToText init")

    @google_dependency_required
    def init_speech_to_text_client(
        self,
        service_credentials_file: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Cloud Speech to Text client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            speech.SpeechClient,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def recognize(
        self,
        audio_file_uri,
        encoding: str = "FLAC",
        language_code: str = "en_US",
        audio_channel_count=2,
    ):
        """Recognize text in the audio file

        :param audio_file_uri: Google Cloud Storage URI
        :return: recognized texts
        """
        # flac or wav, does not require encoding type
        service = self._get_service(self._service_name)
        audio = speech.types.RecognitionAudio(  # pylint: disable=E1101
            uri=audio_file_uri
        )
        if encoding and encoding not in self._encodings:
            encoding = self._encodings["UNSPECIFIED"]
        config = speech.types.RecognitionConfig(  # pylint: disable=E1101
            encoding=encoding,
            language_code=language_code,
            audio_channel_count=audio_channel_count,
            use_enhanced=True,
        )
        rec = service.recognize(config=config, audio=audio)
        return rec.results


class ServiceStorage(GoogleBase):
    """Class for Google Cloud Storage API
     and Google Cloud Storage JSON API

    You will have to grant the appropriate permissions to the
    service account you are using to authenticate with
    @google-cloud/storage. The IAM page in the console is here:
    https://console.cloud.google.com/iam-admin/iam/project

    Link to `Google Storage PyPI`_ page.

    .. _Google Storage PyPI: https://pypi.org/project/google-cloud-storage/
    """

    _service_name = "storage"

    def __init__(self) -> None:
        self.services.append(self._service_name)
        self.logger.debug("ServiceStorage init")

    @google_dependency_required
    def init_storage_client(
        self,
        service_credentials_file: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Cloud Storage client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._init_service(
            storage.Client,
            self._service_name,
            service_credentials_file,
            use_robocloud_vault,
        )

    @google_dependency_required
    def create_bucket(self, bucket_name: str):
        """Create Google Cloud Storage bucket

        :param bucket_name: name as string
        :return: bucket
        """
        service = self._get_service(self._service_name)
        bucket = service.create_bucket(bucket_name)
        return bucket

    @google_dependency_required
    def delete_bucket(self, bucket_name: str):
        """Delete Google Cloud Storage bucket

        Bucket needs to be empty before it can be deleted.

        :param bucket_name: name as string
        """
        bucket = self.get_bucket(bucket_name)
        try:
            bucket.delete()
        except Exception as e:
            raise ValueError("The bucket you tried to delete was not empty") from e

    @google_dependency_required
    def get_bucket(self, bucket_name: str):
        """Get Google Cloud Storage bucket

        :param bucket_name: name as string
        :return: bucket
        """
        if not bucket_name:
            raise KeyError("bucket_name is required for kw: get_bucket")
        service = self._get_service(self._service_name)
        bucket = service.get_bucket(bucket_name)
        return bucket

    @google_dependency_required
    def list_buckets(self) -> list:
        """List Google Cloud Storage buckets

        :return: list of buckets
        """
        service = self._get_service(self._service_name)
        buckets = list(service.list_buckets())
        return buckets

    @google_dependency_required
    def delete_files(self, bucket_name: str, files: Any):
        """Delete files in the bucket

        Files need to be object name in the bucket.

        :param bucket_name: name as string
        :param files: single file, list of files or
            comma separated list of files
        :return: list of files which could not be deleted,
            or True if all were deleted
        """
        if not bucket_name or not files:
            raise KeyError("bucket_name and files are required for kw: delete_files")
        if not isinstance(files, list):
            files = files.split(",")
        bucket = self.get_bucket(bucket_name)
        notfound = []
        for filename in files:
            filename = filename.strip()
            blob = bucket.get_blob(filename)
            if blob:
                blob.delete()
            else:
                notfound.append(filename)
        return notfound if len(notfound) > 0 else True

    @google_dependency_required
    def list_files(self, bucket_name: str):
        """List files in the bucket

        :param bucket_name: name as string
        :return: list of object names in the bucket
        """
        if not bucket_name:
            raise KeyError("bucket_name is required for kw: list_files")
        bucket = self.get_bucket(bucket_name)
        all_blobs = bucket.list_blobs()
        return sorted(blob.name for blob in all_blobs)

    @google_dependency_required
    def upload_file(self, bucket_name: str, filename: str, target_name: str):
        """Upload a file into a bucket

        :param bucket_name: name as string
        :param filename: filepath to upload file
        :param target_name: target object name
        """
        if not bucket_name or not filename or not target_name:
            raise KeyError(
                "bucket_name, filename and target_name are required for kw: upload_file"
            )
        bucket = self.get_bucket(bucket_name)
        blob = bucket.blob(target_name)
        with open(filename, "rb") as f:
            blob.upload_from_file(f)

    @google_dependency_required
    def upload_files(self, bucket_name: str, files: dict):
        """Upload files into a bucket

        Example `files`:
        files = {"mytestimg": "image1.png", "mydoc": "google.pdf"}

        :param bucket_name: name as string
        :param files: dictionary of object names and filepaths
        """
        if not bucket_name or not files:
            raise KeyError("bucket_name and files are required for kw: upload_files")
        if not isinstance(files, dict):
            raise ValueError("files needs to be an dictionary")
        bucket = self.get_bucket(bucket_name)
        for target_name, filename in files.items():
            blob = bucket.blob(target_name)
            blob.upload_from_filename(filename)

    @google_dependency_required
    def download_files(self, bucket_name: str, files: Any):
        """Download files from a bucket

        Example `files`:
        files = {"mytestimg": "image1.png", "mydoc": "google.pdf

        :param bucket_name: name as string
        :param files: list of object names or dictionary of
            object names and target files
        :return: list of files which could not be downloaded, or
            True if all were downloaded
        """
        if isinstance(files, str):
            files = files.split(",")
        bucket = self.get_bucket(bucket_name)
        notfound = []
        if isinstance(files, dict):
            for object_name, filename in files.items():
                blob = bucket.get_blob(object_name)
                if blob:
                    with open(filename, "wb") as f:
                        blob.download_to_file(f)
                        self.logger.info(
                            "Downloaded object %s from Google to filepath %s",
                            object_name,
                            filename,
                        )
                else:
                    notfound.append(object_name)
        else:
            for filename in files:
                filename = filename.strip()
                blob = bucket.get_blob(filename)
                if blob:
                    with open(filename, "wb") as f:
                        blob.download_to_file(f)
                        self.logger.info(
                            "Downloaded object %s from Google to filepath %s",
                            filename,
                            filename,
                        )
                else:
                    notfound.append(filename)
        return notfound if len(notfound) > 0 else True


class ServiceSheets(GoogleBase):
    """Class for Google Sheets API

    You will have to grant the appropriate permissions to the
    service account you are using to authenticate with
    Google Sheets API. The IAM page in the console is here:
    https://console.cloud.google.com/iam-admin/iam/project

    For more information about Google Sheets API link_.

    .. _link: https://developers.google.com/sheets/api/quickstart/python
    """

    _service_name = "sheets"

    def __init__(self) -> None:
        self.services.append(self._service_name)

    @google_dependency_required
    def init_sheets_client(
        self,
        service_credentials_file: str = None,
        use_robocloud_vault: bool = False,
    ) -> None:
        """Initialize Google Sheets client

        :param service_credentials_file: filepath to credentials JSON
        :param use_robocloud_vault: use json stored into `Robocloud Vault`
        """
        self._scopes = [
            "https://www.googleapis.com/auth/drive",
            "https://www.googleapis.com/auth/drive.file",
            "https://www.googleapis.com/auth/spreadsheets",
        ]
        service_account_file = None
        service = None
        if use_robocloud_vault:
            service_account_file = self._get_service_account_from_robocloud()
        elif service_credentials_file:
            service_account_file = service_credentials_file
        try:
            credentials = service_account.Credentials.from_service_account_file(
                service_account_file, scopes=self._scopes
            )
            service = discovery.build(
                "sheets", "v4", credentials=credentials, cache_discovery=False
            )
        except OSError as e:
            raise AssertionError from e
        finally:
            if use_robocloud_vault:
                os.remove(service_account_file)

        self._set_service(self._service_name, service)

    @google_dependency_required
    def create_sheet(self, title: str) -> str:
        """Create empty sheet with a title

        :param title: name as string
        :return: created `sheet_id`
        """
        if not title:
            raise KeyError("title is required for kw: create_sheet")

        service = self._get_service(self._service_name)
        data = {"properties": {"title": title}}
        spreadsheet = (
            service.spreadsheets().create(body=data, fields="spreadsheetId").execute()
        )
        return spreadsheet.get("spreadsheetId")

    @google_dependency_required
    def insert_values(
        self,
        sheet_id: str,
        sheet_range: str,
        values: list,
        major_dimension: str = "COLUMNS",
        value_input_option: str = "USER_ENTERED",
    ) -> None:
        """Insert values into sheet cells

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :param values: list of values to insert into sheet
        :param major_dimension: major dimension of the values, default `COLUMNS`
        :param value_input_option: controls whether input strings are parsed or not,
                                   default `USER_ENTERED`
        """
        if not sheet_id or not sheet_range:
            raise KeyError(
                "sheet_id and sheet_range are required for kw: insert_values"
            )
        if not values:
            raise ValueError("Please provide list of values to insert into sheet")

        service = self._get_service(self._service_name)

        datavalues = []
        for val in values:
            datavalues.append([val])
        resource = {"majorDimension": major_dimension, "values": datavalues}
        service.spreadsheets().values().append(
            spreadsheetId=sheet_id,
            range=sheet_range,
            body=resource,
            valueInputOption=value_input_option,
        ).execute()

    def get_values(
        self,
        sheet_id: str,
        sheet_range: str,
        value_render_option: str = "UNFORMATTED_VALUE",
        datetime_render_option: str = "FORMATTED_STRING",
    ) -> list:
        """Get values from the range in the sheet

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :param value_render_option: how values should be represented
                                    in the output defaults to "UNFORMATTED_VALUE"
        :param datetime_render_option: how dates, times, and durations should be
                                       represented in the output, defaults to "FORMATTED_STRING"
        """  # noqa: E501
        service = self._get_service(self._service_name)
        values = (
            service.spreadsheets()
            .values()
            .get(
                spreadsheetId=sheet_id,
                range=sheet_range,
                valueRenderOption=value_render_option,
                dateTimeRenderOption=datetime_render_option,
            )
            .execute()
        )
        return values

    def clear_values(self, sheet_id: str, sheet_range: str) -> None:
        """Clear cell values for range of cells within a sheet

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        """
        service = self._get_service(self._service_name)
        service.spreadsheets().values().clear(
            spreadsheetId=sheet_id,
            range=sheet_range,
        ).execute()


class ServiceAppsScript(GoogleBase):
    """Class for Google Apps Script API

    **Note:** The Apps Script API does not work with _service accounts_

    Following steps are needed to authenticate and use the service:

    1. enable Apps Script API in the Cloud Platform project (GCP)
    2. create OAuth credentials so API can be authorized (download ``credentials.json``
       which is needed to initialize service)
    3. the Google Script needs to be linked to Cloud Platform project number
    4. Google Script needs to have necessary OAuth scopes to access app
       which is being scripted
    5. necessary authentication scopes and credentials.json are needed
       to initialize service and run scripts

    For more information about Google Apps Script API link_.

    .. _link: https://developers.google.com/apps-script/api
    """

    _service_name = "apps_script"

    def __init__(self) -> None:
        self.services.append(self._service_name)

    @google_dependency_required
    def init_apps_script_client(
        self,
        credentials_file: str = "credentials.json",
        token_file: str = "oauth.token",
        use_robocloud_vault: bool = False,
        scopes: list = None,
        save_token: bool = False,
    ) -> None:
        """Initialize Google Apps Script client

        :param credentials_file: filepath to credentials JSON
        :param token_file: filepath to OAuth token file
        :param use_robocloud_vault: if `True` the token is read from Robocloud
        :param scopes: authenticated scopes, for example. ['forms', 'spreadsheets']
        :param save_token: set to `True` if token should be saved to local file
        """
        self._scopes = ["https://www.googleapis.com/auth/script.projects"]
        credentials = self._get_credentials_with_oauth_token(
            use_robocloud_vault, token_file, credentials_file, scopes, save_token
        )

        service = build("script", "v1", credentials=credentials, cache_discovery=False)
        self._set_service(self._service_name, service)

    def run_script(self, script_id: str, function_name: str, parameters: dict) -> None:
        """Run the Google Apps Script

        :param script_id: Google Script identifier
        :param function_name: name of the script function
        :param parameters: script function parameters as a dictionary
        :raises AssertionError: thrown when Google Script returns errors

        Example:

        .. code-block:: robotframework

            &{params}=    Create Dictionary  formid=aaad4232  formvalues=1,2,3
            ${response}=  Run Script    abc21397283712da  submit_form   ${params}
            Log Many   ${response}
        """
        request = {
            "function": function_name,
            "parameters": [parameters],
        }
        service = self._get_service(self._service_name)
        response = (
            service.scripts()
            .run(
                body=request,
                scriptId=script_id,
            )
            .execute()
        )
        if "error" in response.keys():
            raise AssertionError(response["error"])
        return response


class ServiceDrive(GoogleBase):
    """Class for Google Drive API

    **Note:** The Drive API does not work with _service accounts_

    Following steps are needed to authenticate and use the service:

    1. enable Drive API in the Cloud Platform project (GCP)
    2. create OAuth credentials so API can be authorized (download ``credentials.json``
       which is needed to initialize service)
    3. necessary authentication scopes and credentials.json are needed
       to initialize service

    For more information about Google Drive API link_.

    .. _link: https://developers.google.com/drive/api
    """

    _service_name = "drive"

    def __init__(self) -> None:
        self.services.append(self._service_name)

    @google_dependency_required
    def init_drive_client(
        self,
        credentials_file: str = "credentials.json",
        token_file: str = "oauth.token",
        use_robocloud_vault: bool = False,
        scopes: list = None,
        save_token: bool = False,
    ) -> None:
        """Initialize Google Drive client

        :param credentials_file: filepath to credentials JSON
        :param token_file: filepath to OAuth token file
        :param use_robocloud_vault: if `True` the token is read from Robocloud
        :param scopes: authenticated scopes, for example. ['forms', 'spreadsheets']
        :param save_token: set to `True` if token should be saved to local file
        """
        self._scopes = [
            "https://www.googleapis.com/auth/drive.appdata",
            "https://www.googleapis.com/auth/drive.file",
            "https://www.googleapis.com/auth/drive.install",
        ]
        credentials = self._get_credentials_with_oauth_token(
            use_robocloud_vault, token_file, credentials_file, scopes, save_token
        )
        service = build("drive", "v3", credentials=credentials, cache_discovery=False)
        self.logger.info(service)
        self._set_service(self._service_name, service)

    def drive_upload_file(
        self,
        filename: str = None,
        folder: str = None,
        overwrite: bool = False,
        make_dir: bool = False,
    ) -> dict:
        """Upload files into Drive

        :param filename: name of the file to upload
        :param folder: target folder for upload
        :param overwrite: set to `True` if already existing file should be overwritten
        :param make_dir: set to `True` if folder should be created if it does not exist
        :raises GoogleDriveError: if target_folder does not exist or
         trying to upload directory
        :return: uploaded file id

        Example:

        .. code-block:: robotframework

            ${file1_id}=  Drive Upload File   data.json  # Upload file to drive root
            ${file2_id}=  Drive Upload File   newdata.json  new_folder  make_dir=True
            ${file3_id}=  Drive Upload File   data.json  overwrite=True
        """
        service = self._get_service(self._service_name)

        folder_id = self.drive_get_folder_id(folder)
        if folder_id is None and make_dir:
            folder_id = self.drive_create_directory(folder)
        if folder_id is None:
            raise GoogleDriveError(
                "Target folder '%s' does not exist or could not be created" % folder
            )

        filepath = Path(filename)
        if filepath.is_dir():
            raise GoogleDriveError(
                "The '%s' is a directory and can not be uploaded" % filename
            )
        elif not filepath.is_file():
            raise GoogleDriveError("Filename '%s' does not exist" % filename)

        query_string = f"name = '{filename}' and '{folder_id}' in parents"
        target_file = self.drive_search_files(query=query_string, recurse=True)
        file_metadata = {
            "name": filepath.name,
            "parents": [folder_id],
            "mimeType": "*/*",
        }
        media = MediaFileUpload(filepath.absolute(), mimetype="*/*", resumable=True)
        if len(target_file) == 1 and overwrite:
            self.logger.info("Overwriting file '%s' with new content", filename)
            file = (
                service.files()
                .update(fileId=target_file[0]["id"], media_body=media, fields="id")
                .execute()
            )
            return file.get("id", None)
        elif len(target_file) == 1 and not overwrite:
            self.logger.info("Not uploading new copy of file '%s'", filename)
            return None
        elif len(target_file) > 1:
            self.logger.warning(
                "Drive already contains '%s' copies of file '%s'. Not uploading again."
                % (len(target_file), filename)
            )
            return None
        else:
            file = (
                service.files()
                .create(body=file_metadata, media_body=media, fields="id")
                .execute()
            )
            return file.get("id", None)

    def _download_with_fileobject(self, file_object):
        service = self._get_service(self._service_name)
        request = service.files().get_media(fileId=file_object["id"])
        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()
        fh.seek(0)
        with open(file_object["name"], "wb") as f:
            # pylint: disable=protected-access
            shutil.copyfileobj(fh, f, length=downloader._total_size)

    def drive_download_files(
        self,
        file_dict: dict = None,
        query: str = None,
        limit: int = None,
        timeout: float = None,
    ):
        """Download files specified by file dictionary or query string

        Parameters `start`, `limit` and `timeout` are used only when
        downloading files defined by `query` parameter.

        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target files, defaults to None
        :param start: start index from which to start download files
        :param limit: maximum amount of files that are downloaded, defaults to None
        :param timeout: maximum allowed time in seconds for download process
        :return: list of downloaded files

        Example:

        .. code-block:: robotframework

            ${files}=    Drive Search Files    query=name contains '.json'
            FOR    ${f}    IN    @{files}
                Run Keyword If  ${f}[size] < 2000  Drive Download Files  file_dict=${f}
            END

            ${folder_id}=   Drive Get Folder Id   datafolder
            Drive Download Files  query=name contains '.json' and '${folder_id}' in parents  recurse=True
        """  # noqa: E501
        if query:
            filelist = self.drive_search_files(query)
            files_downloaded = []
            start_time = time.time()

            for f in filelist:
                self._download_with_fileobject(f)
                current_time = time.time()
                files_downloaded.append(f["name"])
                if limit and len(files_downloaded) >= limit:
                    self.logger.info(
                        "Drive download limit %s reached. Stopping the download.", limit
                    )
                    break
                if timeout and (current_time - start_time) > float(timeout):
                    self.logger.info(
                        "Drive download timeout %s seconds reached. "
                        "Stopping the download.",
                        timeout,
                    )
                    break
            return files_downloaded

        if file_dict:
            self._download_with_fileobject(file_dict)
            return [file_dict]
        return None

    def drive_update_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        action: Update = Update.star,
        multiple_ok: bool = False,
    ):
        """Update file specified by id, file dictionary or query string

        Possible actions:
        - star
        - unstar
        - trash
        - untrash

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param action: update action, default star file
        :param multiple_ok: set to `True` if it is ok to perform update
         on more than 1 file
        :return: number of updated files

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Drive Get Folder Id   datafolder
            ${updated}=    Drive Update File   query=name contains '.json' and '${folder_id}' in parents
            ...            action=star
            ...            multiple_ok=True
        """  # noqa: E501
        target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)
        action = to_action(action)
        update_count = 0
        for tf in target_files:
            self._drive_files_update(tf, action)
            update_count += 1
        return update_count

    def _get_target_file(self, file_id, file_dict, query, multiple_ok):
        target_files = []
        if file_id:
            target_files.append(file_id)
        elif file_dict:
            target_files.append(file_dict.get("id", None))
        else:
            files = self.drive_search_files(query, recurse=True)
            target_files = [tf.get("id", None) for tf in files]
            if not multiple_ok and len(target_files) > 1:
                raise GoogleDriveError(
                    "expected search to match 1 file, but it matched %s files"
                    % len(files)
                )

        return target_files

    def _drive_files_update(self, file_id: str, action: Update):
        service = self._get_service(self._service_name)
        body = None
        if action == Update.trash:
            body = {"trashed": True}
        elif action == Update.untrash:
            body = {"trashed": False}
        elif action == Update.star:
            body = {"starred": True}
        elif action == Update.unstar:
            body = {"starred": False}
        else:
            # TODO: mypy should handle enum exhaustivity validation
            raise ValueError(f"Unsupported update action: {action}")
        updated_file = service.files().update(fileId=file_id, body=body).execute()
        return updated_file

    def drive_delete_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        multiple_ok: bool = False,
    ):
        """Delete file specified by id, file dictionary or query string

        Note. Be extra careful when calling this keyword!

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target file, needs to match 1 file
         unless parameter `multiple_ok` is set to `True`
        :param multiple_ok: set to `True` if it is ok to perform delete
         on more than 1 file
        :return: how many files where deleted

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Drive Get Folder Id   datafolder
            ${deleted}=    Drive Delete File   query=name contains '.json' and '${folder_id}' in parents
            ...            multiple_ok=True
        """  # noqa: E501
        service = self._get_service(self._service_name)
        target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)

        delete_count = 0
        for tf in target_files:
            service.files().delete(fileId=tf).execute()
            delete_count += 1
        return delete_count

    def drive_get_folder_id(self, folder: str = None):
        """Get file id for the folder

        :param folder: name of the folder to identify, by default returns drive's
         `root` folder id
        :return: file id of the folder

        Example:

        .. code-block:: robotframework

            ${root_id}=    Drive Get Folder Id   # returns Drive root folder id
            ${folder_id}=  Drive Get Folder Id  subdir
        """
        service = self._get_service(self._service_name)
        mime_folder_type = "application/vnd.google-apps.folder"
        folder_id = None
        if folder is None:
            file = service.files().get(fileId="root", fields="id").execute()
            folder_id = file.get("id", None)
        else:
            query_string = f"name = '{folder}' AND mimeType = '{mime_folder_type}'"
            folders = self.drive_search_files(query=query_string, recurse=True)
            if len(folders) == 1:
                folder_id = folders[0].get("id", None)
            else:
                self.logger.info(
                    "Found '%s' directories with name '%s'", (len(folders), folder)
                )
        return folder_id

    def drive_move_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        folder: str = None,
        folder_id: str = None,
        multiple_ok: bool = False,
    ):
        """Move file specified by id, file dictionary or query string into target folder

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param folder: name of the folder to move file into, is by default drive's
         `root` folder id
        :param folder_id: id of the folder to move file into, if set the `folder`
         parameter is ignored
        :param multiple_ok: if `True` then moving more than 1 file
        :return: list of file ids
        :raises GoogleDriveError: if there are no files to move or
         target folder can't be found

        Example:

        .. code-block:: robotframework

            ${source_id}=  Drive Get Folder Id  sourcefolder
            ${query}=      Set Variable  name contains '.json' and '${sourceid}' in parents
            ${files}=      Drive Move File  query=${query}  folder=target_folder  multiple_ok=True
        """  # noqa: E501
        result_files = []
        service = self._get_service(self._service_name)
        target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)
        if len(target_files) == 0:
            raise GoogleDriveError("Did not find any files to move")
        if folder_id:
            target_parent = folder_id
        else:
            target_parent = self.drive_get_folder_id(folder)
        if target_parent is None:
            raise GoogleDriveError(
                "Unable to find target folder: '%s'" % (folder if folder else "root")
            )
        for tf in target_files:
            file = service.files().get(fileId=tf, fields="parents").execute()
            previous_parents = ",".join(file.get("parents"))
            result_file = (
                service.files()
                .update(
                    fileId=tf,
                    addParents=target_parent,
                    removeParents=previous_parents,
                    fields="id, parents",
                )
                .execute()
            )
            result_files.append(result_file)
        return result_files

    def drive_search_files(
        self, query: str = None, recurse: bool = False, folder_name: str = None
    ) -> list:
        """Search Google Drive for files matching query string

        :param query: search string, defaults to None which means that all files
         and folders are returned
        :param recurse: set to `True` if search should recursive
        :param folder_name: search files in this directory
        :raises GoogleDriveError: if there is a request error
        :return: list of files

        Example:

        .. code-block:: robotframework

            ${files}=  Drive Search Files   query=name contains 'hello'
            ${files}=  Drive Search Files   query=modifiedTime > '2020-06-04T12:00:00'
            ${files}=  Drive Search Files   query=mimeType contains 'image/' or mimeType contains 'video/'
            ${files}=  Drive Search Files   query=name contains '.yaml'  recurse=True
            ${files}=  Drive Search Files   query=name contains '.yaml'  folder_name=datadirectory
        """  # noqa: E501
        service = self._get_service(self._service_name)
        page_token = None
        filelist = []
        parameters = {"fields": "nextPageToken, files", "spaces": "drive", "q": ""}
        query_string = [query] if query else []

        if not recurse:
            folder_id = (
                "root" if not folder_name else self.drive_get_folder_id(folder_name)
            )
            if folder_id is None:
                return []
            query_string.append(f"'{folder_id}' in parents")

        parameters["q"] = " and ".join(query_string)

        while True:
            parameters["pageToken"] = page_token
            try:
                response = service.files().list(**parameters).execute()
            except HttpError as e:
                raise GoogleDriveError(str(e)) from e
            for file in response.get("files", []):
                filesize = file.get("size")
                filelist.append(
                    {
                        "name": file.get("name"),
                        "id": file.get("id"),
                        "size": int(filesize) if filesize else None,
                        "kind": file.get("kind"),
                        "parents": file.get("parents"),
                        "starred": file.get("starred"),
                        "trashed": file.get("trashed"),
                        "shared": file.get("shared"),
                        "mimeType": file.get("mimeType"),
                        "spaces": file.get("spaces", None),
                        "exportLinks": file.get("exportLinks"),
                        "createdTime": file.get("createdTime"),
                        "modifiedTime": file.get("modifiedTime"),
                    }
                )
            page_token = response.get("nextPageToken", None)
            if page_token is None:
                break
        return filelist

    def drive_create_directory(self, folder: str = None):
        """Create new directory to Google Drive

        :param folder: name for the new directory
        :raises GoogleDriveError: if folder name is not given
        :return: created file id
        """
        if not folder or len(folder) == 0:
            raise GoogleDriveError("Can't create Drive directory with empty name")

        folder_id = self.drive_get_folder_id(folder)
        if folder_id:
            self.logger.info("Folder '%s' already exists. Not creating new one.")
            return None

        service = self._get_service(self._service_name)
        file_metadata = {
            "name": folder,
            "mimeType": "application/vnd.google-apps.folder",
        }

        added_folder = service.files().create(body=file_metadata, fields="id").execute()
        return added_folder

    def drive_export_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        target_file: str = None,
        mimetype: str = "application/pdf",
    ):
        """Export Google Drive file using Drive export links

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param target_file: name for the exported file
        :param mimetype: export mimetype, defaults to "application/pdf"

        Example:

        .. code-block:: robotframework

            ${files}=  Drive Search Files  query=name contains 'my example worksheet'
            Drive Export File  file_dict=${files}[0]
        """
        if target_file is None or len(target_file) == 0:
            raise AttributeError("The target_file is required parameter for export")
        if file_id is None and file_dict is None:
            raise AttributeError("Either file_id or file_dict is required for export")
        service = self._get_service(self._service_name)
        target_files = self._get_target_file(file_id, file_dict, None, False)
        if len(target_files) != 1:
            raise ValueError("Did not find the Google Drive file to export")
        request = service.files().export(fileId=target_files[0], mimeType=mimetype)

        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()

        fh.seek(0)
        filepath = Path(target_file).absolute()
        with open(filepath, "wb") as f:
            shutil.copyfileobj(fh, f, length=131072)


class Google(  # pylint: disable=R0901
    ServiceVision,
    ServiceNaturalLanguage,
    ServiceVideoIntelligence,
    ServiceTranslation,
    ServiceTextToSpeech,
    ServiceSpeechToText,
    ServiceStorage,
    ServiceSheets,
    ServiceAppsScript,
    ServiceDrive,
):
    """`Google` is a library for operating with Google API endpoints.

    Usage requires the following steps:

    - Create a GCP project
    - Create a service account key file (JSON) and save it to a place the robot
      can use it
    - Enable APIs
    - Install rpaframework[google]

    **Google authentication**

    Authentication for Google is set with `service credentials JSON file` which can be given to the library
    in three different ways or with `credentials.json`, which is used for OAuth authentication.

    Methods when using service credentials:

    - Method 1 as environment variables, ``GOOGLE_APPLICATION_CREDENTIALS`` with path to JSON file.
    - Method 2 as keyword parameter to ``Init Storage Client`` for example.
    - Method 3 as Robocloud vault secret. The vault name and secret key name needs to be given in library init
      or with keyword ``Set Robocloud Vault``. Secret value should contain JSON file contents.

    Method 1. service credentials using environment variable

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            # NO parameters for Vision Client, expecting to get JSON
            # with GOOGLE_APPLICATION_CREDENTIALS environment variable
            Init Vision Client

    Method 2. service credentials with keyword parameter

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            Init Speech To Text Client  /path/to/service_credentials.json

    Method 3. setting Robocloud Vault in the library init

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google
        ...       robocloud_vault_name=googlecloud
        ...       robocloud_vault_secret_key=servicecreds

        *** Tasks ***
        Init Google services
            Init Storage Client   use_robocloud_vault=${TRUE}

    Method 3. setting Robocloud Vault with keyword

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            Set Robocloud Vault   vault_name=googlecloud  vault_secret_key=servicecreds
            Init Storage Client   use_robocloud_vault=${TRUE}

    Method when using OAuth credentials.json:

    The Google Apps Script and Google Drive services are authenticated using this method.

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Variables ***
        @{SCRIPT_SCOPES}     forms   spreadsheets

        *** Tasks ***
        Init Google OAuth services
            Init Apps Script Client   /path/to/credentials.json   ${SCRIPT_SCOPES}

    **Creating and using OAuth token file**

    The token file can be created using `credentials.json` by running command:

    ``rpa-google-oauth --service drive`` or
    ``rpa-google-oauth --scopes drive.appdata,drive.file,drive.install``

    This will start web based authentication process, which outputs the token at the end.
    Token could be stored into ``Robocorp Vault`` where it needs to be in variable ``google-oauth``.

    Example Vault content.

    .. code-block:: json

        "googlecloud": {
            "oauth-token": "gANfd123321aabeedYsc"
        }

    Using the Vault.

    .. code-block:: robotframework

        *** Keywords ***
        Set up Google Drive authentication
            Set Robocloud Vault    vault_name=googlecloud
            Init Drive Client    use_robocloud_vault=True


    **Requirements**

    Due to number of dependencies related to Google Cloud services this library has been set as
    an optional package for ``rpaframework``.

    This can be installed by opting in to the `google` dependency:

    ``pip install rpaframework[google]``

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Variables ***
        ${SERVICE CREDENTIALS}    ${/}path${/}to${/}service_credentials.json
        ${BUCKET_NAME}            testbucket12213123123

        *** Tasks ***
        Upload a file into a new storage bucket
            [Setup]   Init Storage Client   ${SERVICE CREDENTIALS}
            Create Bucket    ${BUCKET_NAME}
            Upload File      ${BUCKET_NAME}   ${/}path${/}to${/}file.pdf  myfile.pdf
            @{files}         List Files   ${BUCKET_NAME}
            FOR   ${file}  IN   @{files}
                Log  ${file}
            END

    **Python**

    .. code-block:: python

        from RPA.Cloud.Google import Google

        library = Google
        service_credentials = '/path/to/service_credentials.json'

        library.init_vision_client(service_credentials)
        library.init_text_to_speech(service_credentials)

        response = library.detect_text('imagefile.png', 'result.json')
        library.synthesize_speech('I want this said aloud', target_file='said.mp3')
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self, robocloud_vault_name: str = None, robocloud_vault_secret_key: str = None
    ):
        self.set_robocloud_vault(robocloud_vault_name, robocloud_vault_secret_key)
        self.logger = logging.getLogger(__name__)
        ServiceVision.__init__(self)
        ServiceNaturalLanguage.__init__(self)
        ServiceVideoIntelligence.__init__(self)
        ServiceTranslation.__init__(self)
        ServiceTextToSpeech.__init__(self)
        ServiceSpeechToText.__init__(self)
        ServiceStorage.__init__(self)
        ServiceSheets.__init__(self)
        ServiceAppsScript.__init__(self)
        ServiceDrive.__init__(self)
        self.logger.info("Google library initialized")



import cgi
import json
import logging
import os
import threading
import time
from collections import OrderedDict
from datetime import datetime
from http.server import (  # pylint: disable=no-name-in-module
    BaseHTTPRequestHandler,
    HTTPServer,
)
from itertools import count
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import unquote_plus

import requests
from RPA.Browser.Selenium import Selenium

try:
    from importlib import import_module
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources


class Handler(BaseHTTPRequestHandler):
    """Server request handler class"""

    # pylint: disable=unused-argument, signature-differs
    def log_message(self, *args, **kwargs):
        return

    # pylint: disable=unused-argument, signature-differs
    def log_request(self, *args, **kwargs):
        pass

    def _set_response(self, response_code=200, headertype="json"):
        self.send_response(response_code)
        if headertype == "json":
            self.send_header("Content-type", "application/json")
        else:
            self.send_header("Content-type", "text/html")
        self.send_header("Access-Control-Allow-Origin", "*")
        self.end_headers()

    def import_styles(self):
        inline_styles = "<head><style>"
        stylesheet_filepath = Path(self.server.stylepath)
        with open(stylesheet_filepath, "r") as styles:
            inline_styles += styles.read()
        inline_styles += "</style></head>"
        return inline_styles

    def get_radiobutton(self, item):
        formhtml = ""
        if "label" in item:
            formhtml += f"<p>{item['label']}</p>"
        for option in item["options"]:
            checkedvalue = ""
            if "default" in item and item["default"] == option:
                checkedvalue = " checked"
            formhtml += f"""<input type=\"radio\" id=\"{option}\" name=\"{item['id']}\"
                            value="{option}"{checkedvalue}>
                            <label for=\"{option}\">{option}</label><br>"""
        return formhtml

    def get_checkbox(self, item):
        formhtml = ""
        if "label" in item:
            formhtml += f"<p>{item['label']}</p>"
        idx = 1
        for option in item["options"]:
            checkedvalue = ""
            if "default" in item and item["default"] == option:
                checkedvalue = " checked"
            formhtml += f"""<input type=\"checkbox\" id=\"{item['id']}{idx}\"
                            name=\"{item['id']}{idx}\" value="{option}" {checkedvalue}>
                            <label for=\"{item['id']}{idx}\">{option}</label><br>"""
            idx += 1
        return formhtml

    def get_dropdown(self, item):
        formhtml = (
            f"<label for=\"{item['id']}\">{item['label']}</label><br>"
            f"<select name=\"{item['id']}\" name=\"{item['id']}\"><br>"
        )
        for option in item["options"]:
            selected = ""
            if "default" in item and item["default"] == option:
                selected = " selected"
            formhtml += f'<option name="{option}"{selected}>{option}</option>'
        formhtml += "</select><br>"
        return formhtml

    def get_submit(self, item):
        formhtml = ""
        for button in item["buttons"]:
            formhtml += f"""<input type=\"submit\" name=\"{item['name']}\"
                        value=\"{button}\">"""
        return formhtml + "<br>"

    def get_textarea(self, item):
        defaulttext = item["default"] if "default" in item else ""
        return f"""<textarea name=\"{item['name']}\" rows=\"{item['rows']}\"
                    cols=\"{item['cols']}\">{defaulttext}</textarea><br>"""

    def get_textinput(self, item):
        value = item["value"] if "value" in item else ""
        return (
            f"<label for=\"{item['name']}\">{item['label']}</label><br>"
            f"<input type=\"text\" name=\"{item['name']}\" value=\"{value}\"><br>"
        )

    def get_hiddeninput(self, item):
        return (
            f"<input type=\"hidden\" name=\"{item['name']}\""
            f"value=\"{item['value']}\"><br>"
        )

    def get_fileinput(self, item):
        accept_filetypes = (
            f"accept=\"{item['filetypes']}\"" if "filetypes" in item.keys() else ""
        )
        formhtml = (
            f"<label for=\"{item['name']}\">{item['label']}</label><br>"
            f"<input type=\"file\" id=\"{item['id']}\" "
            f"name=\"{item['name']}\" {accept_filetypes} multiple><br>"
        )
        if "target_directory" in item.keys():
            formhtml += (
                f"<input type='hidden' name='target_directory'"
                f" value=\"{item['target_directory']}\">"
            )
        return formhtml

    def get_title(self, item):
        return f"<h3>{item['value']}</h3>"

    def get_text(self, item):
        return f"<p>{item['value']}</p>"

    def create_form(self, message):
        has_submit = False
        formhtml = "<head>"
        formhtml += self.import_styles()
        formhtml += "</head>"
        formhtml += '<form action="formresponsehandling" method="post"'
        formhtml += ' enctype="multipart/form-data">'
        for item in message["form"]:
            dom_func = getattr(self, f"get_{item['type']}", "")
            formhtml += dom_func(item)
            if item["type"] == "submit":
                has_submit = True
        if not has_submit:
            formhtml += "<input type='submit' value='Submit'>"
        formhtml += "</form></body>"

        self.server.formhtml = formhtml.encode("utf-8")

    def do_HEAD(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_POST(self):
        if "formspec" in self.path:
            # read the message and convert it into a python dictionary
            length = int(self.headers.get("content-length"))
            message = json.loads(self.rfile.read(length), object_pairs_hook=OrderedDict)
            self.server.formresponse = None
            self.create_form(message)
            self._set_response()
        elif "formresponsehandling" in self.path:
            length = int(self.headers.get("Content-length", 0))
            form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={
                    "REQUEST_METHOD": "POST",
                    "CONTENT_TYPE": self.headers["Content-Type"],
                },
            )
            files = []
            target_directory = os.getenv("ROBOT_ROOT", str(Path.cwd()))

            response = {}
            for field in form.list or ():
                if field.filename:
                    files.append(field)
                elif field.name == "target_directory":
                    target_directory = Path(field.value).resolve()
                else:
                    response[str(field.name)] = unquote_plus(str(field.value))

            for f in files:
                field_name = str(f.name)
                os.makedirs(target_directory, exist_ok=True)
                filepath = str(target_directory / Path(f.filename))
                with open(filepath, "wb") as fw:
                    fw.write(f.file.read())
                if field_name in response.keys():
                    response[field_name].append(filepath)
                else:
                    response[field_name] = [filepath]

            self.server.formresponse = response
            self._set_response()
        else:
            self._set_response(404)

    def do_GET(self):
        if self.path.endswith("favicon.ico"):
            return

        if self.path.endswith("requestresponse"):
            if self.server.formresponse is not None:
                self._set_response(200, "json")
                self.wfile.write(json.dumps(self.server.formresponse).encode("utf-8"))
                self.server.formresponse = None
            else:
                self._set_response(304, "json")
        elif self.path == "/":
            if self.server.formhtml:
                self._set_response(200, "html")
                self.wfile.write(self.server.formhtml)
            else:
                self._set_response(404, "html")
        elif self.path.endswith(".html"):
            self._set_response(200, "html")
            if self.path == "/":
                filename = "./index.html"
            else:
                filename = "./" + self.path
            with open(filename, "rb") as fh:
                html = fh.read()
                self.wfile.write(html)
        else:
            self._set_response(404, "html")


class Dialogs:
    """The `Dialogs` library provides a way to ask for user input during executions
    through HTML forms. Form elements can be built with library keywords or they can
    be defined in a static JSON file.

    **How the library works**

    The main keyword of the library is ``Request Response`` which works as follows:

    1. It starts an HTTP server in the background
    2. The HTML form is generated either according to a JSON file or the
       keywords called during the task
    3. It opens a browser and shows the created form (The browser is opened with
       the ``Open Available Browser`` keyword from the ``RPA.Browser.Selenium`` library)
    4. Once the form is filled and submitted by the user, the server will process
       the response and extract the field values, which in turn are returned by the keyword
    5. In the end, the browser is closed and the HTTP server is stopped

    ``Request Response`` can be invoked in two ways:

    1. Without any parameters. This means that form shown is the one created
       by other library keywords. If no form elements have been added with
       keywords then the form will contain just one submit button. Form building
       must be started with the keyword ``Create Form``.
    2. Giving a path to a JSON file (using the parameter **formspec**) which
       specifies the elements that form should include.

    The keyword has optional parameters to specify form window **width** and **height**.
    The default size is 600px wide and 1000px high.

    **Setting library arguments**

    Library has arguments ``server_port`` and ``stylesheet``. The ``server_port`` argument
    takes integer value, which defines port where HTTP server will be run. By default port is 8105.
    The ``stylesheet`` can be used to point CSS file, which will be used to modify style of form,
    which is shown to the user. Defaults to built-in Robocorp stylesheet.

    **Supported element types**

    As a bare minimum, the form is displayed with a submit button when the
    ``Request Response`` keyword is called.

    The supported input elements and their corresponding HTML tags are:

    - form (``<form>``)
    - title (``<h3>``)
    - text (``<p>``)
    - radiobutton  (``<input type='radio'>``)
    - checkbox (``<input type='checkbox'>``)
    - dropdown (``<select>``)
    - textarea (``<textarea>``)
    - textinput (``<input type='text'>``)
    - fileinput (``<input type='file'>``)
    - hiddeninput (``<input type='hidden'>``)
    - submit (``<input type='submit'>``)

    **About file types**

    The ``Add File Input`` keyword has parameter ``filetypes``. Parameter sets filter
    for file types that can be uploaded via element. Parameter can be set to ``filetypes=${EMPTY}``
    to accept all files. Multiple types are separated with comma ``filetypes=image/jpeg,image/png``.

    Some common filetypes:

    - image/* (all image types)
    - audio/* (all audio types)
    - video/* (all video types)
    - application/pdf (PDFs)
    - application/vnd.ms-excel (.xls, .xlsx)

    The list of all possible `MIME-types <http://www.iana.org/assignments/media-types/media-types.xhtml>`_.

    **Examples**

    **Robot Framework**

    Examples of creating forms through keywords and a JSON file:

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Dialogs

        *** Keywords ***
        Ask Question From User By Form Built With Keywords
            Create Form     questions
            Add Text Input  label=What is your name?  name=username
            &{response}=    Request Response
            Log             Username is "${response}[username]"

        Ask Question From User By Form Specified With JSON
            &{response}=    Request Response  /path/to/myform.json
            Log             Username is "${response}[username]"

    **Python**

    The library can also be used inside Python:

    .. code-block:: python

        from RPA.Dialogs import Dialogs

        def ask_question_from_user(question, attribute):
            d = Dialogs()
            d.create_form('questions')
            d.add_text_input(label=question, name=attribute)
            response = request_response()
            return response

        response = ask_question_from_user('What is your name ?', 'username')
        print(f"Username is '{response['username']}'")
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, server_port: int = 8105, stylesheet: str = None):
        """The dialogs library can be initialized to a custom
        port and with a custom stylesheet for dialogs.

        :param server_port: HTTP server port, defaults to 8105
        :param stylesheet: defaults to built-in Robocorp stylesheet
        """
        self.logger = logging.getLogger(__name__)
        self.server_port = int(server_port)
        self.server_address = f"http://localhost:{self.server_port}"
        self.server = None
        self.thread = None

        self.custom_form = None
        self.stylesheet = stylesheet

        if self.stylesheet is None:
            includes = import_module("RPA.includes")
            with pkg_resources.path(includes, "dialog_styles.css") as path:
                self.stylesheet = path

    def _start_server_thread(self):
        """Start a server which will server form html and
        handles form post.
        """
        # Already running
        if self.thread and self.thread.is_alive():
            return

        # Previous server thread has died
        if self.thread and not self.thread.is_alive():
            self.server.shutdown()
            self.server.server_close()
            self.thread.join(timeout=10)

        self.thread = threading.Thread(name="form_server", target=self._run_server)
        self.thread.daemon = True
        self.thread.start()

    def _run_server(self):
        self.logger.info("Starting server at port %d", self.server_port)

        self.server = HTTPServer(("", self.server_port), Handler)
        self.server.allow_reuse_address = True
        self.server.stylepath = self.stylesheet
        self.server.formresponse = None
        self.server.formhtml = None

        self.server.serve_forever()

    def create_form(self, title: str = None) -> None:
        """Create new form

        :param title: form title, defaults to None

        Example:

        .. code-block:: robotframework

            Create Form     # form title will be "Requesting response"
            Create Form     title=User Confirmation Form

        """
        self.custom_form = OrderedDict()
        self.custom_form["form"] = list()
        if title:
            self.add_title(title)
        self.add_hidden_input(
            name="dialogs_form_creation_date",
            value=datetime.now(),
        )

    def add_title(self, title: str) -> None:
        """Add h3 element into form

        :param title: text for the element

        Example:

        .. code-block:: robotframework

            Create Form     # default form title will be "Requesting response"
            Add Title       User Confirmation Form

        """
        element = {"type": "title", "value": title}
        self.custom_form["form"].append(element)

    def add_text_input(self, label: str, name: str, value: str = None) -> None:
        """Add text input element

        :param label: input element label
        :param name: input element name attribute
        :param value: input element value attribute

        Example:

        .. code-block:: robotframework

            Create Form
            Add Text Input   what is your firstname ?  fname   value=Mika

        """
        element = {"type": "textinput", "label": label, "name": name, "value": value}
        self.custom_form["form"].append(element)

    def add_hidden_input(self, name: str, value: str) -> None:
        """Add hidden input element

        :param name: input element name attribute
        :param value: input element value attribute

        Example:

        .. code-block:: robotframework

            Create Form
            ${uuid}   Evaluate  str(uuid.uuid4())
            Add Hidden Input    form-id   ${uuid}

        """
        element = {
            "type": "hiddeninput",
            "name": name,
            "value": str(value),
        }
        self.custom_form["form"].append(element)

    def add_file_input(
        self,
        label: str,
        element_id: str,
        name: str,
        filetypes: str,
        target_directory: str = None,
    ) -> None:
        """Add text input element

        :param label: input element label
        :param element_id: hidden element id attribute
        :param name: input element name attribute
        :param filetypes: accepted filetypes for the file upload
        :param target_directory: where to save uploaded files to

        Read more of the filetypes in the library documentation.

        Example:

        .. code-block:: robotframework

            Create Form
            Add File Input  label=Attachment
            ...             element_id=attachment
            ...             name=attachment
            ...             filetypes=${EMPTY}         # Accept all files
            ...             target_directory=${CURDIR}${/}output

            Add File Input  label=Contract
            ...             element_id=contract
            ...             name=contract
            ...             filetypes=application/pdf  # Accept only PDFs
            ...             target_directory=${CURDIR}${/}output

        """
        element = {
            "type": "fileinput",
            "label": label,
            "name": name,
            "id": element_id,
            "filetypes": filetypes,
        }
        if target_directory:
            element["target_directory"] = target_directory
        self.custom_form["form"].append(element)

    def add_dropdown(
        self, label: str, element_id: str, options: Any, default: str = None
    ) -> None:
        """Add dropdown element

        :param label: dropdown element label
        :param element_id: dropdown element id attribute
        :param options: values for the dropdown
        :param default: dropdown selected value, defaults to None

        Example:

        .. code-block:: robotframework

            Create Form
            Add Dropdown  label=Select task type
            ...           element_id=tasktype
            ...           options=buy,sell,rent
            ...           default=buy

        """
        if not isinstance(options, list):
            options = options.split(",")
        element = {
            "type": "dropdown",
            "label": label,
            "id": element_id,
            "options": options,
        }
        if default:
            element["default"] = default
        self.custom_form["form"].append(element)

    def add_submit(self, name: str, buttons: str) -> None:
        """Add submit element

        :param name: element name attribute
        :param buttons: list of buttons

        Example:

        .. code-block:: robotframework

            Create Form
            Add Submit    name=direction-to-go  buttons=left,right

        """
        if not isinstance(buttons, list):
            buttons = buttons.split(",")
        element = {"type": "submit", "name": name, "buttons": buttons}
        self.custom_form["form"].append(element)

    def add_radio_buttons(
        self, element_id: str, options: str, default: str = None
    ) -> None:
        """Add radio button element

        :param element_id: radio button element identifier
        :param options: values for the radio button
        :param default: radio button selected value, defaults to None

        Example:

        .. code-block:: robotframework

            Create Form
            Add Radio Button   element_id=drone  buttons=Jim,Robert  default=Robert

        """
        if not isinstance(options, list):
            options = options.split(",")
        element = {
            "type": "radiobutton",
            "id": element_id,
            "options": options,
        }
        if default is not None:
            element["default"] = default
        self.custom_form["form"].append(element)

    def add_checkbox(
        self, label: str, element_id: str, options: str, default: str = None
    ) -> None:
        """Add checkbox element

        :param label: check box element label
        :param element_id: check box element identifier
        :param options: values for the check box
        :param default: check box selected value, defaults to None

        Example:

        .. code-block:: robotframework

            Create Form
            Add Checkbox    label=Select your colors
            ...             element_id=colors
            ...             options=green,red,blue,yellow
            ...             default=blue

        """
        if not isinstance(options, list):
            options = options.split(",")
        element = {
            "type": "checkbox",
            "label": label,
            "id": element_id,
            "options": options,
        }
        if default is not None:
            element["default"] = default
        self.custom_form["form"].append(element)

    def add_textarea(
        self, name: str, rows: int = 5, cols: int = 40, default: str = None
    ) -> None:
        """Add textarea element

        :param name: textarea element name
        :param rows: number of rows for the area, defaults to 5
        :param cols: numnber of columns for the area, defaults to 40
        :param default: prefilled text for the area, defaults to None

        Example:

        .. code-block:: robotframework

            Create Form
            Add Textarea       name=feedback  default=enter feedback here
            Add Textarea       name=texts  rows=40   cols=80

        """
        element = {
            "type": "textarea",
            "name": name,
            "rows": rows,
            "cols": cols,
        }
        if default is not None:
            element["default"] = default
        self.custom_form["form"].append(element)

    def add_text(self, value: str) -> None:
        """Add text paragraph element

        :param value: text for the element

        Example.

        .. code-block:: robotframework

            Create Form
            Add Text       ${form_guidance_text}

        """
        element = {
            "type": "text",
            "value": value,
        }
        self.custom_form["form"].append(element)

    def request_response(
        self,
        formspec: str = None,
        window_width: int = 600,
        window_height: int = 1000,
        timeout: Optional[int] = None,
    ) -> Dict:
        """Start server and show form. Waits for user response.

        :param formspec: form json specification file, defaults to None
        :param window_width: window width in pixels, defaults to 600
        :param window_height: window height in pixels, defaults to 1000
        :param timeout: optional time to wait for response, in seconds
        :return: form response

        Example:

        .. code-block:: robotframework

            Create Form    ${CURDIR}/${/}myform.json
            &{response}    Request Response

        """
        self._start_server_thread()

        if self.custom_form is None:
            self.create_form("Requesting response")

        if formspec:
            formdata = open(formspec, "rb")
        else:
            formdata = json.dumps(self.custom_form)

        for attempt in count():
            try:
                response = requests.post(
                    f"{self.server_address}/formspec",
                    data=formdata,
                    headers={
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                    },
                    timeout=5,
                )
                response.raise_for_status()
                break
            except Exception as exc:  # pylint: disable=broad-except
                if not self.thread.is_alive() or attempt >= 5:
                    raise RuntimeError("Failed to start input server") from exc
                time.sleep(1)

        return self._wait_response(window_width, window_height, timeout)

    def _wait_response(
        self, window_width: int, window_height: int, timeout: Optional[int]
    ) -> Dict:
        """Open a browser to the created form and wait for the user
        to submit it.
        """
        browser = Selenium()

        def is_browser_open():
            try:
                return bool(browser.driver.current_window_handle)
            except Exception:  # pylint: disable=broad-except
                return False

        try:
            browser.open_available_browser(self.server_address)
            browser.set_window_size(window_width, window_height)

            start_time = time.time()
            while True:
                response = requests.get(
                    f"{self.server_address}/requestresponse",
                    headers={"Prefer": "wait=120"},
                )

                if response.status_code == 200:
                    return response.json()
                elif response.status_code != 304:
                    response.raise_for_status()
                elif not is_browser_open():
                    raise RuntimeError("Browser closed by user")
                elif timeout and time.time() > start_time + int(timeout):
                    raise RuntimeError("No response within timeout")
                else:
                    time.sleep(1)
        finally:
            browser.close_browser()



from datetime import datetime
from fnmatch import fnmatch
import logging
import os
import os.path
from pathlib import Path
import tarfile
from typing import Any
import zipfile


def convert_date(timestamp):
    if isinstance(timestamp, tuple):
        d = datetime(
            year=timestamp[0],
            month=timestamp[1],
            day=timestamp[2],
            hour=timestamp[3],
            minute=timestamp[4],
            second=timestamp[5],
        )
    else:
        d = datetime.utcfromtimestamp(timestamp)
    formatted_date = d.strftime("%d.%m.%Y %H:%M:%S")
    return formatted_date


def list_files_in_directory(folder, recursive=False, include=None, exclude=None):
    filelist = []
    for rootdir, _, files in os.walk(folder):
        for file in files:
            archive_absolute = os.path.join(rootdir, file)
            archive_relative = rootdir.replace(folder, "")
            archive_relative = os.path.join(archive_relative, file)
            if include and not fnmatch(archive_relative, include):
                continue
            if exclude and fnmatch(archive_relative, exclude):
                continue
            filelist.append((archive_absolute, archive_relative))
        if not recursive:
            break
    return filelist


class Archive:
    """`Archive` is a library for operating with ZIP and TAR packages.

    **Examples**

    .. code-block:: robotframework

       *** Settings ***
       Library  RPA.Archive

       *** Tasks ***
       Creating a ZIP archive
          Archive Folder With ZIP   ${CURDIR}${/}tasks  tasks.zip   recursive=True  include=*.robot  exclude=/.*
          @{files}                  List Archive             tasks.zip
          FOR  ${file}  IN  ${files}
             Log  ${file}
          END
          Add To Archive            .${/}..${/}missing.robot  tasks.zip
          &{info}                   Get Archive Info


    .. code-block:: python

        from RPA.Archive import Archive

        lib = Archive()
        lib.archive_folder_with_tar('./tasks', 'tasks.tar', recursive=True)
        files = lib.list_archive('tasks.tar')
        for file in files:
           print(file)
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def archive_folder_with_zip(
        self,
        folder: str,
        archive_name: str,
        recursive: bool = False,
        include: str = None,
        exclude: str = None,
        compression: str = "stored",
    ) -> None:
        # pylint: disable=C0301
        """Create a zip archive of a folder

        :param folder: name of the folder to archive
        :param archive_name: filename of the archive
        :param recursive: should sub directories be included, defaults is False
        :param include: define file pattern to include in the package, defaults to None (means all files)
        :param exclude: define file pattern to exclude from the package, defaults is None
        :param compression: type of package compression method, defaults to "stored"

        This keyword creates an ZIP archive of a local folder. By default subdirectories are not
        included, but they can included using `recursive` argument.

        To include only certain files, like TXT files, the argument `include` can be used.
        Similarly to exclude certain file, like dotfiles, the argument `exclude` can be used.

        Compression methods:

        - stored, default
        - deflated
        - bzip2
        - lzma

        Example:

        .. code-block:: robotframework

            Archive Folder With Zip  ${CURDIR}${/}documents  mydocs.zip
            Archive Folder With Zip  ${CURDIR}${/}tasks      robottasks.zip   include=*.robot
            Archive Folder With Zip  ${CURDIR}${/}tasks      no_dotfiles.zip  exclude=/.*
            Archive Folder With Zip  ${CURDIR}${/}documents  documents.zip    recursive=True
            Archive Folder With Zip  ${CURDIR}               packagelzma.zip  compression=lzma
            Archive Folder With Zip  ${CURDIR}               bzipped.zip      compression=bzip2
        """  # noqa: E501
        if compression == "stored":
            comp_method = zipfile.ZIP_STORED
        elif compression == "deflated":
            comp_method = zipfile.ZIP_DEFLATED
        elif compression == "bzip2":
            comp_method = zipfile.ZIP_BZIP2
        elif compression == "lzma":
            comp_method = zipfile.ZIP_LZMA
        else:
            raise ValueError("Unknown compression method")

        filelist = list_files_in_directory(folder, recursive, include, exclude)
        if len(filelist) == 0:
            raise ValueError("No files found to archive")
        zip_archive = zipfile.ZipFile(
            file=archive_name,
            mode="w",
            compression=comp_method,
        )
        for archive_absolute, archive_relative in filelist:
            zip_archive.write(archive_absolute, arcname=archive_relative)
        zip_archive.close()

    def archive_folder_with_tar(
        self,
        folder: str,
        archive_name: str,
        recursive: bool = False,
        include: str = None,
        exclude: str = None,
    ) -> None:
        """Create a tar/tar.gz archive of a folder

        :param folder: name of the folder to archive
        :param archive_name: filename of the archive
        :param recursive: should sub directories be included, defaults is False
        :param include: define file pattern to include in the package,
            by default all files are included
        :param exclude: define file pattern to exclude from the package

        This keyword creates an TAR or TAR.GZ archive of a local folder. Type of archive
        is determined by the file extension. By default subdirectories are not
        included, but they can included using `recursive` argument.

        To include only certain files, like TXT files, the argument `include` can be used.
        Similarly to exclude certain file, like dotfiles, the argument `exclude` can be used.

        Example:

        .. code-block:: robotframework

            Archive Folder With TAR  ${CURDIR}${/}documents  documents.tar
            Archive Folder With TAR  ${CURDIR}${/}tasks      tasks.tar.gz   include=*.robot
            Archive Folder With TAR  ${CURDIR}${/}tasks      tasks.tar      exclude=/.*
            Archive Folder With TAR  ${CURDIR}${/}documents  documents.tar  recursive=True
        """  # noqa: E501
        filelist = list_files_in_directory(folder, recursive, include, exclude)
        if len(filelist) == 0:
            raise ValueError("No files found to archive")
        tar_archive = tarfile.TarFile(archive_name, "w")
        for archive_absolute, archive_relative in filelist:
            tar_archive.add(archive_absolute, arcname=archive_relative)
        tar_archive.close()

    def add_to_archive(
        self,
        files: Any,
        archive_name: str,
        folder: str = None,
    ) -> None:
        """Add file(s) to the archive

        :param files: name of the file, or list of files, to add
        :param archive_name: filename of the archive
        :param folder: name of the folder for added file,
            relative path in the archive

        This keyword adds file or list of files into existing archive. Files
        can be added to archive structure with relative path using argument `folder`.

        Example:

        .. code-block:: robotframework

            Add To Archive  extrafile.txt  myfiles.zip
            Add To Archive  stat.png       archive.tar.gz  images
            @{files}        Create List    filename1.txt   filename2.txt
            Add To Archive  ${files}       files.tar
        """
        files_to_add = []

        if isinstance(files, list):
            for filename in files:
                file_relative = Path(filename).name
                if folder:
                    file_relative = folder + os.path.sep + file_relative
                files_to_add.append((filename, file_relative))
        else:
            file_relative = Path(files).name
            if folder:
                file_relative = folder + os.path.sep + file_relative
            files_to_add.append((files, file_relative))

        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "a") as zip_archive:
                for filename, file_relative in files_to_add:
                    zip_archive.write(filename, arcname=file_relative)
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "a") as tar_archive:
                for file_absolute, file_relative in files_to_add:
                    tar_archive.add(file_absolute, arcname=file_relative)

    def list_archive(self, archive_name: str) -> list:
        """List files in an archive

        :param archive_name: filename of the archive

        Returns list of file, where each file in a list is a dictionary
        with following attributes:

        - name
        - size
        - mtime
        - last modification time in format `%d.%m.%Y %H:%M:%S`

        Example:

        .. code-block:: robotframework

            @{files}   List Archive    myfiles.zip
            FOR  ${file}  IN   ${files}
                Log  ${file}[filename]
                Log  ${file}[size]
                Log  ${file}[mtime]
            END
        """
        filelist = []
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                members = f.infolist()
                for memb in members:
                    filelist.append(
                        {
                            "filename": memb.filename,
                            "size": memb.file_size,
                            "mtime": memb.date_time,
                            "modified": convert_date(memb.date_time),
                        }
                    )
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "r") as f:
                members = f.getmembers()
                for memb in members:
                    filelist.append(
                        {
                            "name": memb.name,
                            "size": memb.size,
                            "mtime": memb.mtime,
                            "modified": convert_date(memb.mtime),
                        }
                    )
        return filelist

    def get_archive_info(self, archive_name: str) -> dict:
        """Get information about the archive

        :param archive_name: filename of the archive

        Returns following file attributes in a dictionary:

        - filename
        - filemode
        - size
        - mtime
        - last modification time in format `%d.%m.%Y %H:%M:%S`

        Example:

        .. code-block:: robotframework

            &{archiveinfo}   Get Archive Info    myfiles.zip
        """
        archive_info = None
        st = os.stat(archive_name)
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                archive_info = {
                    "filename": f.filename,
                    "mode": f.mode,
                    "size": st.st_size,
                    "mtime": st.st_mtime,
                    "modified": convert_date(st.st_mtime),
                }
        elif tarfile.is_tarfile(archive_name):
            with tarfile.TarFile(archive_name, "r") as f:
                archive_info = {
                    "filename": f.name,
                    "mode": f.mode,
                    "size": st.st_size,
                    "mtime": st.st_mtime,
                    "modified": convert_date(st.st_mtime),
                }
        return archive_info

    def extract_archive(
        self, archive_name: str, path: str = None, members: Any = None
    ) -> None:
        """Extract files from archive into local directory

        :param archive_name: filename of the archive
        :param path: filepath to extract file into, default is current working directory
        :param members: list of files to extract from, by default
            all files in archive are extracted

        This keyword supports extracting files from zip, tar and tar.gz archives.

        By default file is extracted into current working directory, but `path` argument
        can be set to define extraction path.

        Example:

        .. code-block:: robotframework

            Extract Archive    myfiles.zip   ${CURDIR}${/}extracted
            @{files}           Create List   filename1.txt    filename2.txt
            Extract Archive    archive.tar   C:${/}myfiles${/}  ${files}
        """  # noqa: E501
        root = Path(path) if path else Path.cwd()
        if members and not isinstance(members, list):
            members = [members]
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as f:
                if members:
                    f.extractall(path=root, members=members)
                else:
                    f.extractall(path=root)
        elif tarfile.is_tarfile(archive_name):
            members = map(tarfile.TarInfo, members) if members else None
            with tarfile.open(archive_name, "r") as f:
                if members:
                    f.extractall(path=root, members=members)
                else:
                    f.extractall(path=root)

    def extract_file_from_archive(
        self, filename: str, archive_name: str, path: str = None
    ):  # pylint: disable=C0301
        """Extract a file from archive into local directory

        :param filename: name of the file to extract
        :param archive_name: filename of the archive
        :param path: filepath to extract file into,
            default is current working directory

        This keyword supports extracting a file from zip, tar and tar.gz archives.

        By default file is extracted into current working directory,
        but `path` argument can be set to define extraction path.

        Example:

        .. code-block:: robotframework

            Extract File From Archive    extrafile.txt   myfiles.zip
            Extract File From Archive    background.png  images.tar.gz  ${CURDIR}${/}extracted
        """  # noqa: E501
        root = Path(path) if path else Path.cwd()
        if zipfile.is_zipfile(archive_name):
            with zipfile.ZipFile(archive_name, "r") as outfile:
                outfile.extract(filename, root)
        elif tarfile.is_tarfile(archive_name):
            with tarfile.open(archive_name, "r") as outfile:
                outfile.extract(filename, root)



import pytest
from pathlib import Path
from RPA.Images import Images, TemplateMatcher, Region, HAS_OPENCV, to_image

IMAGES = Path(__file__).resolve().parent / ".." / "resources" / "images"

# TODO: Create new .png test data which works for pillow matching method
TEMPLATES = (
    ((26, 994, 274, 1115), "locator_Calculator_ctrl_One.jpg"),
    ((279, 994, 527, 1115), "locator_Calculator_ctrl_Two.jpg"),
    ((533, 994, 781, 1115), "locator_Calculator_ctrl_Three.jpg"),
    ((26, 868, 274, 989), "locator_Calculator_ctrl_Four.jpg"),
    ((279, 868, 527, 989), "locator_Calculator_ctrl_Five.jpg"),
    ((533, 868, 781, 989), "locator_Calculator_ctrl_Six.jpg"),
    ((26, 742, 274, 863), "locator_Calculator_ctrl_Seven.jpg"),
    ((279, 742, 527, 863), "locator_Calculator_ctrl_Eight.jpg"),
    ((533, 742, 781, 863), "locator_Calculator_ctrl_Nine.jpg"),
    ((26, 1120, 274, 1241), "locator_Calculator_ctrl_Positive_Negative.jpg"),
    ((279, 1120, 527, 1240), "locator_Calculator_ctrl_Zero.jpg"),
    ((533, 1120, 781, 1240), "locator_Calculator_ctrl_Decimal_Separator.jpg"),
    ((26, 742, 780, 1240), "locator_Calculator_ctrl_Number_pad.jpg"),
    ((786, 616, 1034, 737), "locator_Calculator_ctrl_Divide_by.jpg"),
    ((786, 742, 1034, 863), "locator_Calculator_ctrl_Multiply_by.jpg"),
    ((786, 868, 1034, 988), "locator_Calculator_ctrl_Minus.jpg"),
    ((786, 994, 1034, 1115), "locator_Calculator_ctrl_Plus.jpg"),
    ((786, 1120, 1034, 1240), "locator_Calculator_ctrl_Equals.jpg"),
    ((786, 616, 1034, 1240), "locator_Calculator_ctrl_Standard_operators.jpg"),
    ((26, 616, 274, 737), "locator_Calculator_ctrl_Reciprocal.jpg"),
    ((279, 616, 527, 737), "locator_Calculator_ctrl_Square.jpg"),
    ((533, 616, 781, 737), "locator_Calculator_ctrl_Square_root.jpg"),
    ((26, 616, 780, 737), "locator_Calculator_ctrl_Standard_functions.jpg"),
    ((27, 413, 190, 486), "locator_Calculator_ctrl_Clear_all_memory.jpg"),
    ((195, 413, 358, 486), "locator_Calculator_ctrl_Memory_Recall.jpg"),
    ((364, 413, 528, 486), "locator_Calculator_ctrl_Memory_Add.jpg"),
    ((532, 413, 695, 486), "locator_Calculator_ctrl_Memory_Subtract.jpg"),
    ((701, 413, 865, 486), "locator_Calculator_ctrl_Memory_Store.jpg"),
    ((872, 413, 1035, 486), "locator_Calculator_ctrl_Open_memory_flyout.jpg"),
    ((27, 413, 1035, 486), "locator_Calculator_ctrl_Memory_controls.jpg"),
    ((26, 490, 274, 611), "locator_Calculator_ctrl_Percent.jpg"),
    ((533, 490, 781, 611), "locator_Calculator_ctrl_Clear.jpg"),
    ((279, 490, 527, 611), "locator_Calculator_ctrl_Clear_entry.jpg"),
    ((786, 490, 1034, 611), "locator_Calculator_ctrl_Backspace.jpg"),
    ((26, 490, 1034, 611), "locator_Calculator_ctrl_Display_controls.jpg"),
    ((146, 96, 344, 163), "locator_Calculator_ctrl_Standard_Calculator_mode.jpg"),
    ((16, 81, 116, 181), "locator_Calculator_ctrl_Open_Navigation.jpg"),
    ((944, 81, 1044, 181), "locator_Calculator_ctrl_Open_history_flyout.jpg"),
    ((699, 1, 814, 81), "locator_Calculator_ctrl_Minimize_Calculator.jpg"),
    ((814, 1, 929, 81), "locator_Calculator_ctrl_Maximize_Calculator.jpg"),
    ((374, 81, 474, 181), "locator_Calculator_ctrl_Keep_on_top.jpg"),
    ((929, 1, 1044, 81), "locator_Calculator_ctrl_Close_Calculator.jpg"),
    ((61, 183, 999, 230), "locator_Calculator_ctrl_Expression_is_330__4.jpg"),
    ((16, 230, 1044, 410), "locator_Calculator_ctrl_Display_is_825.jpg"),
)


def get_test_name(param):
    return param[1].split(".")[0]


@pytest.fixture(params=TEMPLATES, ids=get_test_name)
def region_and_template(request):
    yield request.param[0], request.param[1]


@pytest.mark.skipif(not HAS_OPENCV, reason="Test requires opencv support")
def test_find_template(region_and_template):
    region, template = region_and_template
    region = Region(*region)

    library = Images()
    library.matcher = TemplateMatcher(opencv=True)

    matches = library.find_template_in_image(
        image=IMAGES / "source.png", template=IMAGES / template
    )

    assert len(matches) == 1
    match = matches[0]
    assert match.center == region.center


@pytest.mark.skipif(not HAS_OPENCV, reason="Test requires opencv support")
def test_wait_template(region_and_template):
    _, template = region_and_template
    library = Images()
    library.take_screenshot = lambda: to_image(IMAGES / "source.png")

    library.matcher = TemplateMatcher(opencv=True)
    matches = library.wait_template_on_screen(IMAGES / template, timeout=0.52)
    assert len(matches) == 1


@pytest.mark.skip(
    reason="this currently fails because the found template has some offset, at least on multi-monitor setups"
)
def test_screenshot_region_and_find_it():
    library = Images()
    region = Region(0, 0, 100, 100)
    first_capture = library.take_screenshot(region=region)
    find_result = library.find_template_on_screen(first_capture)
    assert region == find_result[0]



import importlib
import logging
import sys

from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from RPA.RobotLogListener import RobotLogListener
from RPA.Tables import Table

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


class Configuration:
    """Class to handle configuration from config files
    and class init"""

    def __init__(self):
        self.configuration = {}
        self.module_name = None

    def parse_arguments(
        self,
        module_name,
        database,
        username,
        password,
        host,
        port,
        charset,
        config_file: str,
    ):
        config = ConfigParser.ConfigParser()
        config.read([config_file])

        self.configuration = {}
        self.module_name = module_name or (
            config.get("default", "module_name")
            if config.has_option("default", "module_name")
            else None
        )
        self.configuration["database"] = database or (
            config.get("default", "database")
            if config.has_option("default", "database")
            else None
        )
        self.configuration["username"] = username or (
            config.get("default", "username")
            if config.has_option("default", "username")
            else None
        )
        self.configuration["password"] = password or (
            config.get("default", "password")
            if config.has_option("default", "password")
            else None
        )
        self.configuration["host"] = host or (
            config.get("default", "host")
            if config.has_option("default", "host")
            else None
        )
        self.configuration["port"] = port or (
            int(config.get("default", "port"))
            if config.has_option("default", "host")
            else None
        )
        self.configuration["charset"] = charset or (
            config.get("default", "charset")
            if config.has_option("default", "charset")
            else None
        )
        return self.module_name, self.configuration

    def get(self, param, default=None):
        return (
            self.configuration[param] if param in self.configuration.keys() else default
        )

    def set_val(self, param, value):
        self.configuration[param] = value

    def all_but_empty(self):
        new_dict = {}
        for key, value in dict(self.configuration).items():
            if value is not None:
                new_dict[key] = value
        return new_dict

    def set_default_port(self, port):
        if (
            "port" not in self.configuration.keys()
            or self.configuration["port"] is None
        ):
            self.configuration["port"] = int(port)

    def get_connection_parameters_as_string(self, conf=None):
        configuration = conf or self.configuration
        parameters = ",".join(
            "{}={}".format(str(k), str(v)) for k, v in configuration.items()
        )
        return "Connecting using : %s.connect(%s)" % (self.module_name, parameters)


class Database:
    """`Database` is a library for handling different database operations.

    All database operations are supported. Keywords `Query` and `Get Rows`
    return values by default in `RPA.Table` format.

    Library is compatible with any Database API Specification 2.0 module.

    References:

    - Database API Specification 2.0 - http://www.python.org/dev/peps/pep-0249/
    - Lists of DB API 2.0 - http://wiki.python.org/moin/DatabaseInterfaces
    - Python Database Programming - http://wiki.python.org/moin/DatabaseProgramming/

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library         RPA.Database

        *** Tasks ***
        Get Orders From Database
            Connect To Database  pymysql  tester  user  password  127.0.0.1
            @{orders}            Query    Select * FROM incoming_orders
            FOR   ${order}  IN  @{orders}
                Handle Order  ${order}
            END

    **Python**

    .. code-block:: python

        from RPA.Database import Database
        from RPA.Robocloud.Secrets import FileSecrets

        filesecrets = FileSecrets("secrets.json")
        secrets = filesecrets.get_secret("databasesecrets")

        db = Database()
        db.connect_to_database('pymysql',
                            secrets["DATABASE"],
                            secrets["USERNAME"],
                            secrets["PASSWORD"],
                            '127.0.0.1'
                            )
        orders = db.query("SELECT * FROM incoming_orders")
        for order in orders:
            print(order)
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self._dbconnection = None
        self.db_api_module_name = None
        self.config = Configuration()
        listener = RobotLogListener()
        listener.register_protected_keywords(["RPA.Database.connect_to_database"])

    # pylint: disable=R0915
    def connect_to_database(  # noqa: C901
        self,
        module_name: str = None,
        database: str = None,
        username: str = None,
        password: str = None,
        host: str = None,
        port: int = None,
        charset: str = None,
        config_file: str = "db.cfg",
    ):
        """Connect to database using DB API 2.0 module.

        :param module_name: database module to use
        :param database: name of the database
        :param username: of the user accessing the database
        :param password: of the user accessing the database
        :param host: SQL server address
        :param port: SQL server port
        :param charset: for example, "utf-8", defaults to None
        :param config_file: location of configuration file, defaults to "db.cfg"

        Example:

        .. code-block:: robotframework

            Connect To Database  pymysql  database  username  password  host  port
            Connect To Database  ${CURDIR}${/}resources${/}dbconfig.cfg

        """
        self.config.parse_arguments(
            module_name, database, username, password, host, port, charset, config_file
        )
        if self.config.module_name in ("excel", "excelrw"):
            self.db_api_module_name = "pyodbc"
            dbmodule = importlib.import_module("pyodbc")
        else:
            self.db_api_module_name = self.config.module_name
            dbmodule = importlib.import_module(self.config.module_name)
        if module_name in ["MySQLdb", "pymysql"]:
            self.config.set_default_port(3306)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                db=self.config.get("database"),
                user=self.config.get("username"),
                passwd=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
                charset=self.config.get("charset"),
            )
        elif module_name == "psycopg2":
            self.config.set_default_port(5432)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                database=self.config.get("database"),
                user=self.config.get("username"),
                password=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
            )
        elif module_name in ("pyodbc", "pypyodbc"):
            self.config.set_default_port(1433)
            self.config.set_val(
                "connect_string",
                "DRIVER={SQL Server};SERVER=%s,%s;DATABASE=%s;UID=%s;PWD=%s"
                % (
                    self.config.get("host"),
                    self.config.get("port"),
                    self.config.get("database"),
                    self.config.get("username"),
                    self.config.get("password"),
                ),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(self.config.get("connect_string"))
        elif module_name == "excel":
            self.config.set_val(
                "connect_string",
                "DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};"
                'DBQ=%s;ReadOnly=1;Extended Properties="Excel 8.0;HDR=YES";)'
                % self.config.get("database"),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                autocommit=True,
            )
        elif module_name == "excelrw":
            self.config.set_val(
                "connect_string",
                "DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};"
                'DBQ=%s;ReadOnly=0;Extended Properties="Excel 8.0;HDR=YES";)'
                % self.config.get("database"),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                autocommit=True,
            )
        elif module_name in ("ibm_db", "ibm_db_dbi"):
            self.config.set_default_port(50000)
            self.config.set_val(
                "connect_string",
                "DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s;"
                % (
                    self.config.get("database"),
                    self.config.get("host"),
                    self.config.get("port"),
                    self.config.get("usernamme"),
                    self.config.get("password"),
                ),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                "",
                "",
            )
        elif module_name == "cx_Oracle":
            self.config.set_default_port(1521)
            oracle_dsn = dbmodule.makedsn(
                host=self.config.get("host"),
                port=self.config.get("port"),
                service_name=self.config.get("database"),
            )
            self.config.set_val("oracle_dsn", oracle_dsn)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                user=self.config.get("username"),
                password=self.config.get("password"),
                dsn=self.config.get("oracle_dsn"),
            )
        elif module_name == "teradata":
            self.config.set_default_port(1025)
            teradata_udaExec = dbmodule.UdaExec(
                appName="RobotFramework", version="1.0", logConsole=False
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = teradata_udaExec.connect(
                method="odbc",
                system=self.config.get("host"),
                database=self.config.get("database"),
                username=self.config.get("username"),
                password=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
            )
        elif module_name == "pymssql":
            self.config.set_default_port(1433)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                server=self.config.get("host"),
                user=self.config.get("username"),
                password=self.config.get("password"),
                database=self.config.get("database"),
                port=self.config.get("port"),
                host=self.config.get("host", "."),
            )
        else:
            conf = self.config.all_but_empty()
            self.logger.info(self.config.get_connection_parameters_as_string(conf))
            self._dbconnection = dbmodule.connect(**conf)

    def call_stored_procedure(self, name, params=None, sanstran=False):
        """Call stored procedure with name and params.

        :param name: procedure name
        :param params: parameters for the procedure as a list, defaults to None
        :param sanstran: run command without an explicit transaction commit or rollback,
         defaults to False

        Example:

        .. code-block:: robotframework

            @{params}     Create List   FirstParam   SecondParam   ThirdParam
            @{results}    Call Stored Procedure   mystpr  ${params}

        """
        params = params or []
        cur = None
        try:
            if self.db_api_module_name == "cx_Oracle":
                cur = self._dbconnection.cursor()
            else:
                cur = self._dbconnection.cursor(as_dict=False)
            PY3K = sys.version_info >= (3, 0)
            if not PY3K:
                name = name.encode("ascii", "ignore")
            cur.callproc(name, params)
            cur.nextset()
            retVal = list()
            for row in cur:
                retVal.append(row)
            if not sanstran:
                self._dbconnection.commit()
            return retVal
        finally:
            if cur:
                if not sanstran:
                    self._dbconnection.rollback()

    def description(self, table):
        """Get description of the SQL table

        :param table: name of the SQL table

        Example:

        .. code-block:: robotframework

            Connect To Database    pymysql  mydb  user  pass  127.0.0.1
            ${db_description}      Description  mytable

        """
        try:
            result = self.query("DESCRIBE %s" % table, as_table=True)
        except Exception as e:
            raise AssertionError(
                "Operation not supported for '%s' type database"
                % self.db_api_module_name
            ) from e
        return result.to_list()

    def disconnect_from_database(self):
        """Close connection to SQL database

        Example:

        .. code-block:: robotframework

            Connect To Database    pymysql  mydb  user  pass  127.0.0.1
            ${result}              Query   Select firstname, lastname FROM table
            Disconnect From Database

        """
        if self._dbconnection:
            self._dbconnection.close()

    # pylint: disable=R0912
    def execute_sql_script(self, filename, sanstran=False):  # noqa: C901
        """Execute content of SQL script as SQL commands.

        :param filename: filepath to SQL script to execute
        :param sanstran: run command without an explicit transaction commit or rollback,
         defaults to False

        Example:

        .. code-block:: robotframework

            Execute SQL Script   script.sql

        """
        sql_script_file = open(filename)

        cur = None
        try:
            cur = self._dbconnection.cursor()
            sqlStatement = ""
            for line in sql_script_file:
                PY3K = sys.version_info >= (3, 0)
                if not PY3K:
                    # spName = spName.encode('ascii', 'ignore')
                    line = line.strip().decode("utf-8")
                if line.startswith("#") or line.startswith("--"):
                    continue

                sql_fragments = line.split(";")
                if len(sql_fragments) == 1:
                    sqlStatement += line + " "
                else:
                    for sqlFragment in sql_fragments:
                        sqlFragment = sqlFragment.strip()
                        if len(sqlFragment) == 0:
                            continue

                        sqlStatement += sqlFragment + " "

                        self.__execute_sql(cur, sqlStatement)
                        sqlStatement = ""

            sqlStatement = sqlStatement.strip()
            if len(sqlStatement) != 0:
                self.__execute_sql(cur, sqlStatement)

            if not sanstran:
                self._dbconnection.commit()
        finally:
            if cur:
                if not sanstran:
                    self._dbconnection.rollback()

    def query(self, statement, assertion=None, sanstran=False, as_table=True):
        """Make a SQL query.

        :param statement: SQL statement to execute
        :param assertion: assert on query result, row_count or columns.
         Works only for SELECT statements Defaults to None.
        :param sanstran: run command without an explicit transaction commit or rollback,
         defaults to False
        :param as_table: if result should be instance of ``Table``, defaults to `True`
         `False` means that return type would be `list`

        Example:

        .. code-block:: robotframework

            @{res}   Query   Select firstname, lastname FROM table
            FOR  ${row}  IN  @{RES}
                Log   ${row}
            END
            @{res}   Query  Select * FROM table  row_count > ${EXPECTED}
            @{res}   Query  Select * FROM table  'arvo' in columns
            @{res}   Query  Select * FROM table  columns == ['id', 'arvo']

        """
        rows = None
        columns = None
        result = None
        cursor = None

        try:
            cursor = self._dbconnection.cursor()
            self.logger.info("Executing : Query  |  %s ", statement)
            result = self.__execute_sql(cursor, statement)

            if statement.lower().startswith("select") or statement.lower().startswith(
                "describe"
            ):
                rows = cursor.fetchall()
                columns = [c[0] for c in cursor.description]
                # pylint: disable=unused-variable
                row_count = len(rows)  # noqa: F841
                if assertion:
                    available_locals = {
                        "row_count": row_count,
                        "columns": columns,
                    }
                    # pylint: disable=W0123
                    valid = eval(assertion, {"__builtins__": None}, available_locals)

                    if not valid:
                        raise AssertionError(
                            "Query assertion %s failed. Facts: %s"
                            % (assertion, available_locals)
                        )
                if as_table:
                    return Table(rows, columns)
                return rows
            else:
                if result is not None:
                    if not sanstran:
                        self._dbconnection.commit()
                if not sanstran:
                    self._dbconnection.commit()
        finally:
            if cursor:
                if not sanstran:
                    self._dbconnection.rollback()
        return result

    def __execute_sql(self, cursor, sqlStatement):
        return cursor.execute(sqlStatement)

    def set_auto_commit(self, autocommit=True):
        """Set database auto commit mode.

        :param autocommit: boolean value for auto commit, defaults to True

        Example:

        .. code-block:: robotframework

            Set Auto Commit             # auto commit is set on
            Set Auto Commit   False     # auto commit is turned off

        """
        self._dbconnection.autocommit = autocommit

    def get_rows(self, table, columns=None, conditions=None, as_table=True):
        """Get rows from table. Columns and conditions can be
        set to filter result.

        :param table: name of the SQL table
        :param columns: name of columns to return, defaults to `None`
         means that all columns are returned
        :param conditions: limiting result by WHERE clause, defaults to `None`
        :param as_table: if result should be instance of ``Table``, defaults to `True`
         `False` means that return type would be `list`

        Example:

        .. code-block:: robotframework

            @{res}   Get Rows  tablename  arvo
            @{res}   Get Rows  tablename  arvo  columns=id,name
            @{res}   Get Rows  tablename  columns=id  conditions=column1='newvalue'
            @{res}   Get Rows  tablename  conditions=column2='updatedvalue'

        """
        columns = columns or "*"
        where_cond = f" WHERE {conditions}" if conditions else ""
        return self.query(
            "SELECT %s FROM %s%s" % (columns, table, where_cond), as_table=as_table
        )

    def get_number_of_rows(self, table, conditions=None):
        """Get number of rows in a table. Conditions can be given
        as arguments for WHERE clause.

        :param table: name of the SQL table
        :param conditions: restrictions for selections, defaults to None

        Example:

        .. code-block:: robotframework

            ${count}   Get Number Of Rows  tablename
            ${count}   Get Number Of Rows  tablename  column1=5 and column2='x'

        """
        where_cond = f" WHERE {conditions}" if conditions else ""
        result = self.query(
            "SELECT COUNT(*) FROM %s%s" % (table, where_cond), as_table=False
        )
        return result[0][0]



import os
import tempfile
from contextlib import contextmanager
from pathlib import Path


RESOURCE_DIR = Path(__file__).resolve().parent / ".." / "resources"


@contextmanager
def temp_filename(content=None):
    """Create temporary file and return filename, delete file afterwards.
    Needs to close file handle, since Windows won't allow multiple
    open handles to the same file.
    """
    with tempfile.NamedTemporaryFile(delete=False) as fd:
        path = fd.name
        if content:
            fd.write(content)

    try:
        yield path
    finally:
        os.unlink(path)
